from datetime import datetime

import pytest
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django_comments.models import CommentFlag
from django_comments_ink import (
    get_comment_reactions_enum,
    get_object_reactions_enum,
)
from django_comments_ink.models import (
    CommentReaction,
    InkComment,
    ObjectReaction,
)
from django_comments_ink.tests.models import Article, Diary


@pytest.mark.django_db
@pytest.fixture
def an_article():
    """Creates an article that can receive comments."""
    return Article.objects.create(
        title="September", slug="september", body="During September..."
    )


@pytest.mark.django_db
@pytest.fixture
def a_diary_entry():
    """Creates a diary entry that can receive comments."""
    return Diary.objects.create(body="About today...")


@pytest.mark.django_db
@pytest.fixture
def an_articles_comment(an_article):
    """Send a comment to the article."""
    article_ct = ContentType.objects.get(app_label="tests", model="article")
    site = Site.objects.get(pk=1)
    return InkComment.objects.create(
        content_type=article_ct,
        object_pk=an_article.pk,
        content_object=an_article,
        site=site,
        comment="First comment to the article.",
        submit_date=datetime.now(),
    )


@pytest.mark.django_db
@pytest.fixture
def an_user():
    """Add a user to the DB."""
    return User.objects.create_user(
        "joe", "joe@example.com", "joepwd", first_name="Joe", last_name="Bloggs"
    )


@pytest.mark.django_db
@pytest.fixture
def an_user_2():
    """Add a user2 to the DB."""
    return User.objects.create_user(
        "alice",
        "alice@example.com",
        "alicepwd",
        first_name="Alice",
        last_name="Bloggs",
    )


@pytest.mark.django_db
@pytest.fixture
def a_comments_reaction(an_articles_comment, an_user):
    """Send a comment reaction to a comment."""
    reaction = get_comment_reactions_enum().LIKE_IT
    cmr = CommentReaction.objects.create(
        reaction=reaction, comment=an_articles_comment, counter=1
    )
    cmr.authors.add(an_user)
    cmr.save()
    return cmr


@pytest.mark.django_db
@pytest.fixture
def a_comments_flag(an_articles_comment, an_user):
    """Send a CommentFlag.SUGGEST_REMOVAL flag to a comment."""
    return CommentFlag.objects.create(
        user=an_user,
        comment=an_articles_comment,
        flag=CommentFlag.SUGGEST_REMOVAL,
    )


@pytest.mark.django_db
@pytest.fixture
def an_object_reaction(a_diary_entry, an_user):
    """Send an object reaction to a diary entry."""
    ctype = ContentType.objects.get_for_model(a_diary_entry)
    site = Site.objects.get(pk=1)
    reaction = get_object_reactions_enum().LIKE_IT
    objr = ObjectReaction.objects.create(
        content_type=ctype,
        object_pk=a_diary_entry.id,
        site=site,
        reaction=reaction,
        counter=1,
    )
    objr.authors.add(an_user)
    objr.save()
    return objr



from django.urls import include, path, re_path
from django_comments.views.comments import comment_done
from django_comments.views.moderation import (
    approve,
    approve_done,
    delete,
    delete_done,
    flag_done,
)
from django_comments_ink import views
from django_comments_ink.conf import settings
from rest_framework.urlpatterns import format_suffix_patterns

urlpatterns = [
    re_path(r"^post/$", views.post, name="comments-ink-post"),
    re_path(r"^posted/$", comment_done, name="comments-comment-done"),
    re_path(r"^sent/$", views.sent, name="comments-ink-sent"),
    re_path(
        r"^confirm/(?P<key>[^/]+)/$", views.confirm, name="comments-ink-confirm"
    ),
    re_path(r"^mute/(?P<key>[^/]+)/$", views.mute, name="comments-ink-mute"),
    re_path(r"^reply/(?P<cid>[\d]+)/$", views.reply, name="comments-ink-reply"),
    # Remap comments-flag to check allow-flagging is enabled.
    re_path(r"^flag/(\d+)/$", views.flag, name="comments-flag"),
    re_path(r"^flagged/$", flag_done, name="comments-flag-done"),
    re_path(r"^delete/(\d+)/$", delete, name="comments-delete"),
    re_path(r"^deleted/$", delete_done, name="comments-delete-done"),
    re_path(r"^approve/(\d+)/$", approve, name="comments-approve"),
    re_path(r"^approved/$", approve_done, name="comments-approve-done"),
    re_path(r"^react/(\d+)/$", views.react, name="comments-ink-react"),
    re_path(r"^reacted/$", views.react_done, name="comments-ink-react-done"),
    re_path(
        r"^list-reacted/(\d+)/([\w\+\-]+)/$",
        views.list_reacted,
        name="comments-ink-list-reacted",
    ),
    re_path(
        r"^react/(\d+)/(\d+)/$",
        views.react_to_object,
        name="comments-ink-react-to-object",
    ),
    re_path(
        r"^list-reacted/(\d+)/(\d+)/([\w\+\-]+)/$",
        views.list_reacted_to_object,
        name="comments-ink-list-reacted-to-object",
    ),
    # Remap comments-url-redirect to add query string params.
    re_path(
        r"^cr/(\d+)/(\d+)/(\d+)/$",
        views.get_inkcomment_url,
        name="comments-url-redirect",
    ),
    # API handlers.
    path(
        "api/",
        include("django_comments_ink.api.urls"),
        {"override_drf_defaults": settings.COMMENTS_INK_OVERRIDE_DRF_DEFAULTS},
    ),
]


urlpatterns = format_suffix_patterns(urlpatterns)



from datetime import datetime

import pytest
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.utils.functional import cached_property
from django_comments_ink import get_model, utils
from django_comments_ink.conf import settings
from django_comments_ink.conf.defaults import COMMENTS_INK_APP_MODEL_OPTIONS
from django_comments_ink.paginator import CommentsPaginator

InkComment = get_model()


@pytest.mark.django_db
def test_send_mail_uses_EmailThread(monkeypatch):
    monkeypatch.setattr(utils.settings, "COMMENTS_INK_THREADED_EMAILS", True)
    utils.send_mail(
        "the subject",
        "the message",
        "helpdesk@example.com",
        ["fulanito@example.com"],
        html="<p>The message.</p>",
    )
    assert utils.mail_sent_queue.get() == True


# ---------------------------------------
send_mail_called = False


def _mocked_send_mail(*args, **kwargs):
    global send_mail_called
    send_mail_called = True


@pytest.mark.django_db
def test_send_mail_uses__send_amil(monkeypatch):
    monkeypatch.setattr(utils, "_send_mail", _mocked_send_mail)
    monkeypatch.setattr(utils.settings, "COMMENTS_INK_THREADED_EMAILS", False)
    utils.send_mail(
        "the subject",
        "the message",
        "helpdesk@example.com",
        ["fulanito@example.com"],
        html="<p>The message.</p>",
    )
    assert send_mail_called == True


# ----------------------------------------------
@pytest.mark.django_db
def test_get_app_model_options_without_args():
    options = utils.get_app_model_options()
    assert options == COMMENTS_INK_APP_MODEL_OPTIONS["default"]


# ----------------------------------------------
only_default_options = {
    "default": {
        "who_can_post": "all",
        "check_input_allowed": "",
        "comment_flagging_enabled": True,
        "comment_reactions_enabled": True,
        "object_reactions_enabled": True,
    }
}


@pytest.mark.django_db
def test_get_app_model_options_without_args_returns_defaults(monkeypatch):
    monkeypatch.setattr(
        utils.settings, "COMMENTS_INK_APP_MODEL_OPTIONS", only_default_options
    )
    app_model_options = utils.get_app_model_options()
    assert app_model_options == only_default_options["default"]


# ----------------------------------------------
default_and_article_options = {
    "default": {
        "who_can_post": "all",
        "check_input_allowed": "",
        "comment_flagging_enabled": True,
        "comment_reactions_enabled": True,
        "object_reactions_enabled": True,
    },
    "tests.article": {
        "comment_flagging_enabled": False,
        "comment_reactions_enabled": False,
        "object_reactions_enabled": False,
    },
}


@pytest.mark.django_db
def test_get_app_model_options_with_comment(an_articles_comment, monkeypatch):
    monkeypatch.setattr(
        utils.settings,
        "COMMENTS_INK_APP_MODEL_OPTIONS",
        default_and_article_options,
    )
    _options = utils.get_app_model_options(an_articles_comment)
    expected_options = default_and_article_options["default"]
    expected_options.update(default_and_article_options["tests.article"])
    assert _options == expected_options


@pytest.mark.django_db
def test_get_app_model_options_with_content_type(monkeypatch):
    monkeypatch.setattr(
        utils.settings,
        "COMMENTS_INK_APP_MODEL_OPTIONS",
        default_and_article_options,
    )
    _options = utils.get_app_model_options(content_type="tests.article")
    expected_options = default_and_article_options["default"]
    expected_options.update(default_and_article_options["tests.article"])
    assert _options == expected_options


@pytest.mark.django_db
def test_get_app_model_options_with_non_existing_content_type(monkeypatch):
    monkeypatch.setattr(
        utils.settings,
        "COMMENTS_INK_APP_MODEL_OPTIONS",
        default_and_article_options,
    )
    _options = utils.get_app_model_options(content_type="tralari.tralara")
    expected_options = default_and_article_options["default"]
    assert _options == expected_options


# ----------------------------------------------
@pytest.mark.django_db
def test_get_user_avatar_for_comment(an_articles_comment):
    gravatar_url = utils.get_user_avatar(an_articles_comment)
    assert gravatar_url.startswith("//www.gravatar.com/avatar/")


# ----------------------------------------------
@pytest.mark.django_db
def test_redirect_to_with_comment(an_articles_comment):
    http_response = utils.redirect_to(an_articles_comment)
    assert http_response.url == an_articles_comment.get_absolute_url()


# ----------------------------------------------
class FakeRequest:
    def __init__(self, cpage):
        self.cpage = cpage

    @property
    def GET(self):
        return {"cpage": self.cpage}


@pytest.mark.django_db
def test_redirect_to_with_request(an_articles_comment):
    request = FakeRequest(cpage=2)
    http_response = utils.redirect_to(an_articles_comment, request)
    assert http_response.url == "/comments/cr/13/1/1/?cpage=2#comment-1"


# ----------------------------------------------
@pytest.mark.django_db
def test_redirect_to_with_page_number(an_articles_comment):
    http_response = utils.redirect_to(an_articles_comment, page_number=2)
    assert http_response.url == "/comments/cr/13/1/1/?cpage=2#comment-1"


# ----------------------------------------------
def create_scenario_1(an_article):
    """Test based on the Example 1 of the paginator.py module."""
    article_ct = ContentType.objects.get(app_label="tests", model="article")
    site = Site.objects.get(pk=1)
    attrs = {
        "content_type": article_ct,
        "object_pk": an_article.pk,
        "content_object": an_article,
        "site": site,
        "comment": f"another comment to article {an_article.pk}",
        "submit_date": datetime.now(),
    }

    # Create 8 comments at level 0.
    for cm_level_0 in range(8):
        InkComment.objects.create(**attrs)

    cmts_level_0 = InkComment.objects.all()
    # Add the following number of child comments to the previous cmts_level_0.
    children_number = [10, 10, 10, 10, 10, 10, 5, 4]
    for index, cmt_level_0 in enumerate(cmts_level_0):
        for child in range(children_number[index]):
            InkComment.objects.create(**attrs, parent_id=cmt_level_0.pk)

    return article_ct


@pytest.mark.django_db
def test_get_comment_page_number_when_page_size_is_0(an_article, monkeypatch):
    article_ct = create_scenario_1(an_article)
    monkeypatch.setattr(utils.settings, "COMMENTS_INK_COMMENTS_PER_PAGE", 0)
    for comment in InkComment.objects.all():
        page_number = utils.get_comment_page_number(
            None, article_ct, an_article.pk, comment.id
        )
        assert page_number == 1


@pytest.mark.django_db
def test_get_comment_page_number(an_article):
    article_ct = create_scenario_1(an_article)
    page_size = 25
    orphans = 10
    assert settings.COMMENTS_INK_MAX_LAST_PAGE_ORPHANS == orphans
    assert settings.COMMENTS_INK_COMMENTS_PER_PAGE == page_size
    queryset = InkComment.objects.all()
    paginator = CommentsPaginator(queryset, page_size, orphans)

    page = paginator.page(1)
    for comment in page.object_list:
        page_number = utils.get_comment_page_number(
            None, article_ct.pk, an_article.pk, comment.id
        )
        assert page_number == 1

    page = paginator.page(2)
    for comment in page.object_list:
        page_number = utils.get_comment_page_number(
            None, article_ct.pk, an_article.pk, comment.id
        )
        assert page_number == 2

    page = paginator.page(3)
    for comment in page.object_list:
        page_number = utils.get_comment_page_number(
            None, article_ct.pk, an_article.pk, comment.id
        )
        assert page_number == 3


# -------------------------------------------
class FakeCommentsPaginator(CommentsPaginator):
    @cached_property
    def num_pages(self):
        return 0


@pytest.mark.django_db
def test_get_comment_page_number_raises_Exception(an_article, monkeypatch):
    article_ct = create_scenario_1(an_article)
    monkeypatch.setattr(utils, "CommentsPaginator", FakeCommentsPaginator)
    page_size = 25
    orphans = 10
    assert settings.COMMENTS_INK_MAX_LAST_PAGE_ORPHANS == orphans
    assert settings.COMMENTS_INK_COMMENTS_PER_PAGE == page_size
    queryset = InkComment.objects.all()
    paginator = CommentsPaginator(queryset, page_size, orphans)

    page = paginator.page(2)
    comment = page.object_list[0]
    with pytest.raises(Exception):
        utils.get_comment_page_number(
            None, article_ct.id, an_article.id, comment.id
        )


@pytest.mark.parametrize(
    "theme_dir, does_exist",
    [
        ("comments/themes/unknown", False),
        ("comments/themes/feedback_in_header", True),
    ],
)
def test_does_theme_dir_exist(theme_dir, does_exist):
    assert utils.does_theme_dir_exist(theme_dir) == does_exist



import hashlib
import logging
import os
import queue as queue
import threading
from urllib.parse import urlencode

from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import EmailMultiAlternatives
from django.db.models import Q
from django.http.response import HttpResponseRedirect
from django.utils.crypto import salted_hmac
from django_comments_ink import get_model
from django_comments_ink.conf import settings
from django_comments_ink.conf.defaults import COMMENTS_INK_APP_MODEL_OPTIONS
from django_comments_ink.paginator import CommentsPaginator
from rest_framework import status
from rest_framework.exceptions import PermissionDenied

logger = logging.getLogger(__name__)

mail_sent_queue = queue.Queue()


class EmailThread(threading.Thread):
    # Idea borrowed from Selwin Ong post:
    # http://ui.co.id/blog/asynchronous-send_mail-in-django
    def __init__(
        self, subject, body, from_email, recipient_list, fail_silently, html
    ):
        self.subject = subject
        self.body = body
        self.recipient_list = recipient_list
        self.from_email = from_email
        self.fail_silently = fail_silently
        self.html = html
        threading.Thread.__init__(self)

    def run(self):
        _send_mail(
            self.subject,
            self.body,
            self.from_email,
            self.recipient_list,
            self.fail_silently,
            self.html,
        )
        mail_sent_queue.put(True)


def _send_mail(
    subject, body, from_email, recipient_list, fail_silently=False, html=None
):
    msg = EmailMultiAlternatives(subject, body, from_email, recipient_list)
    if html:
        msg.attach_alternative(html, "text/html")
    msg.send(fail_silently)


def send_mail(
    subject, body, from_email, recipient_list, fail_silently=False, html=None
):
    if settings.COMMENTS_INK_THREADED_EMAILS:
        EmailThread(
            subject, body, from_email, recipient_list, fail_silently, html
        ).start()
    else:
        _send_mail(
            subject, body, from_email, recipient_list, fail_silently, html
        )


# --------------------------------------------------------------------
def get_app_model_options(comment=None, content_type=None):
    """
    Get the app_model_option from COMMENTS_INK_APP_MODEL_OPTIONS.

    If a comment is given, the content_type is extracted from it. Otherwise,
    the content_type kwarg has to be provided. Checks whether there is a
    matching dictionary for the app_label.model of the content_type, and
    returns it. Otherwise it returns the default from:

        `django_comments_ink.conf.defaults.COMMENTS_INK_APP_MODEL_OPTIONS`.

    """
    defaults_options = dict.copy(COMMENTS_INK_APP_MODEL_OPTIONS)
    settings_options = dict.copy(settings.COMMENTS_INK_APP_MODEL_OPTIONS)
    if defaults_options != settings_options and "default" in settings_options:
        defaults_options["default"].update(settings_options.pop("default"))

    model_app_options = {}
    for app_model in settings_options.keys():
        for k in defaults_options["default"].keys():
            if k not in settings_options[app_model]:
                model_app_options[k] = defaults_options["default"][k]
            else:
                model_app_options[k] = settings_options[app_model][k]

    if comment:
        content_type = ContentType.objects.get_for_model(comment.content_object)
        key = "%s.%s" % (content_type.app_label, content_type.model)
    elif content_type:
        key = content_type
    else:
        return defaults_options["default"]
    try:
        model_app_options.update(settings.COMMENTS_INK_APP_MODEL_OPTIONS[key])
        return model_app_options
    except Exception:
        return model_app_options


option_msgs = {
    "comment_flagging_enabled": {
        "PROD": "This type of comments are not allowed to be flagged.",
        "DEBUG": "Comments posted to instances of '%s.%s' are not "
        "explicitly allowed to be flagged. Check the "
        "COMMENTS_INK_APP_MODEL_OPTIONS setting.",
    },
    "comment_reactions_enabled": {
        "PROD": "This type of comment are not allowed to receive reactions.",
        "DEBUG": "Comments posted to instances of '%s.%s' are not "
        "explicitly allowed to receive reactions. Check the "
        "COMMENTS_INK_APP_MODEL_OPTIONS setting.",
    },
    "object_reactions_enabled": {
        "PROD": "This type of object are not allowed to receive reactions.",
        "DEBUG": "Instances of '%s.%s' are not explicitly allowed to receive "
        "reactions. Check the COMMENTS_INK_APP_MODEL_OPTIONS setting.",
    },
}


def check_option(option, comment=None, content_type=None):
    ret_option = get_app_model_options(
        comment=comment, content_type=content_type
    )[option]
    if not ret_option:
        message = option_msgs[option]["PROD"]
        if settings.DEBUG:
            app_label, model = "", ""
            if comment:
                ct = ContentType.objects.get_for_model(comment.content_object)
                app_label, model = ct.app_label, ct.model
            if content_type:
                app_label, model = content_type.rsplit(".", 1)
            message = option_msgs[option]["DEBUG"] % (app_label, model)
        raise PermissionDenied(detail=message, code=status.HTTP_403_FORBIDDEN)


def check_input_allowed(object):
    """
    This is a generic function called with the object being commented.
    It's defined as the default in 'COMMENTS_INK_APP_MODEL_OPTIONS'.

    If you want to disallow comments input to your 'app.model' instances under
    a given conditions, rewrite this function in your code and modify the
    'COMMENTS_INK_APP_MODEL_OPTIONS' in your settings.
    """
    return True


# --------------------------------------------------------------------
def get_current_site_id(request=None):
    """it's a shortcut"""
    site_id = getattr(settings, "SITE_ID", None)
    if not site_id and request:
        site_id = getattr(get_current_site(request), "pk", 1)  # fallback value
    return site_id


def get_html_id_suffix(object):
    value = "%s" % object.__hash__()
    suffix = salted_hmac(settings.COMMENTS_INK_SALT, value).hexdigest()
    return suffix


def get_user_avatar(comment):
    path = hashlib.md5(comment.user_email.lower().encode("utf-8")).hexdigest()
    param = urlencode({"s": 48})
    return "//www.gravatar.com/avatar/%s?%s&d=identicon" % (path, param)


def get_comment_url(comment, request=None, page_number=None):
    cm_abs_url = comment.get_absolute_url()
    cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
    cpage = request.GET.get(cpage_qs_param, None) if request else page_number
    if cpage:
        hash_pos = cm_abs_url.find("#")
        cm_anchor = cm_abs_url[hash_pos:]
        cm_abs_url = cm_abs_url[:hash_pos]
        url = f"{cm_abs_url}?{cpage_qs_param}={cpage}{cm_anchor}"
        return url
    else:
        return cm_abs_url


def redirect_to(comment, request=None, page_number=None):
    url = get_comment_url(comment, request, page_number)
    return HttpResponseRedirect(url)


def get_comment_page_number(
    request, content_type_id, object_id, comment_id, comments_folded=None
):
    """
    Returns the page number in which the `comment_id` is listed.
    """
    num_orphans = settings.COMMENTS_INK_MAX_LAST_PAGE_ORPHANS
    page_size = settings.COMMENTS_INK_COMMENTS_PER_PAGE
    if page_size == 0:
        return 1  # Pagination is disabled.

    # Replicate the logic in django_comments/templatetags/comments.py,
    # BaseCommentNode.get_queryset method.

    site_id = get_current_site_id(request)

    qs = get_model().objects.filter(
        content_type_id=content_type_id, object_pk=object_id, site__pk=site_id
    )

    # The is_public and is_removed fields are implementation details of the
    # built-in comment model's spam filtering system, so they might not
    # be present on a custom comment model subclass. If they exist, we
    # should filter on them.
    field_names = [f.name for f in get_model()._meta.fields]
    if "is_public" in field_names:
        qs = qs.filter(is_public=True)

    if (
        getattr(settings, "COMMENTS_HIDE_REMOVED", True)
        and "is_removed" in field_names
    ):
        qs = qs.filter(is_removed=False)

    if "user" in field_names:
        qs = qs.select_related("user")

    comment_id = int(comment_id)

    if comments_folded:
        qs = qs.filter(~Q(level__gt=0, thread_id__in=comments_folded))

    ckey_prefix = settings.COMMENTS_INK_CACHE_KEYS["comments_paged"].format(
        ctype_pk=content_type_id, object_pk=object_id, site_id=site_id
    )

    paginator = CommentsPaginator(
        qs,
        page_size,
        orphans=num_orphans,
        comments_folded=comments_folded,
        cache_key_prefix=ckey_prefix,
    )

    for page_number in range(1, paginator.num_pages + 1):
        page = paginator.page(page_number)
        if comment_id in [cm.id for cm in page.object_list]:
            return page_number
    raise Exception("Comment %d not listed in any page." % comment_id)


def does_theme_dir_exist(theme_dir):
    from django.template import engines

    for engine in engines.all():
        for pth in engine.template_dirs:
            if os.path.exists(os.path.join(pth, theme_dir)):
                return True
    return False



import json
from datetime import datetime
from unittest.mock import patch

import django_comments
import pytest
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.db.models.signals import pre_save
from django.test import TestCase as DjangoTestCase
from django.urls import reverse
from django_comments_ink import get_model
from django_comments_ink.api.views import CommentCount, CommentList
from django_comments_ink.conf import settings
from django_comments_ink.models import (
    InkComment,
    publish_or_withhold_on_pre_save,
)
from django_comments_ink.tests.models import Article, MyComment
from django_comments_ink.tests.test_models import (
    thread_test_step_1,
    thread_test_step_2,
    thread_test_step_3,
)
from django_comments_ink.tests.utils import post_comment
from rest_framework.test import APIRequestFactory, APITestCase

app_model_options_mock = {"tests.article": {"who_can_post": "users"}}

InkComment = get_model()


factory = APIRequestFactory()


class CommentCreateTestCase(DjangoTestCase):
    def setUp(self):
        self.patcher = patch("django_comments_ink.views.utils.send_mail")
        self.mock_mailer = self.patcher.start()
        self.article = Article.objects.create(
            title="October", slug="october", body="What I did on October..."
        )
        self.form = django_comments.get_form()(self.article)

    def tearDown(self):
        self.patcher.stop()

    @patch.multiple(
        "django_comments_ink.conf.settings", COMMENTS_INK_CONFIRM_EMAIL=False
    )
    def test_post_returns_201_response(self):
        data = {
            "name": "Bob",
            "email": "fulanito@detal.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal eine kleine...",
            "honeypot": "",
        }
        data.update(self.form.initial)
        response = post_comment(data)
        self.assertEqual(response.status_code, 201)
        data = json.loads(response.rendered_content)
        self.assertTrue("id" in data)
        self.assertEqual(data["id"], 1)  # id of the new created comment.

    def test_post_returns_2xx_response(self):
        data = {
            "name": "Bob",
            "email": "fulanito@detal.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal eine kleine...",
            "honeypot": "",
        }
        data.update(self.form.initial)
        response = post_comment(data)
        self.assertEqual(response.status_code, 204)
        self.assertEqual(self.mock_mailer.call_count, 1)

    def test_post_returns_4xx_response(self):
        # It uses an authenticated user, but the user has no mail address.
        self.user = User.objects.create_user("bob", "", "pwd")
        data = {
            "name": "",
            "email": "",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal eine kleine...",
            "honeypot": "",
        }
        data.update(self.form.initial)
        response = post_comment(data, auth_user=self.user)
        self.assertEqual(response.status_code, 400)
        self.assertTrue("name" in response.data)
        self.assertTrue("email" in response.data)
        self.assertEqual(self.mock_mailer.call_count, 0)

    @patch.multiple(
        "django_comments_ink.conf.settings",
        COMMENTS_INK_APP_MODEL_OPTIONS=app_model_options_mock,
    )
    def test_post_returns_unauthorize_response(self):
        data = {
            "name": "Bob",
            "email": "fulanito@detal.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal eine kleine...",
            "honeypot": "",
        }
        data.update(self.form.initial)
        response = post_comment(data)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.rendered_content, b'"User not authenticated"')
        self.assertEqual(self.mock_mailer.call_count, 0)

    def post_parent_comment(self):
        article_ct = ContentType.objects.get(app_label="tests", model="article")
        site1 = Site.objects.get(pk=1)
        self.cm = InkComment.objects.create(
            content_type=article_ct,
            object_pk=self.article.id,
            content_object=self.article,
            site=site1,
            comment="just a testing comment",
            submit_date=datetime.now(),
        )

    @patch.multiple(
        "django_comments_ink.conf.settings",
        COMMENTS_INK_MAX_THREAD_LEVEL=0,
        COMMENTS_INK_CONFIRM_EMAIL=False,
    )
    def test_post_reply_to_exceeds_max_thread_level_returns_400_code(self):
        self.assertEqual(settings.COMMENTS_INK_MAX_THREAD_LEVEL, 0)
        self.assertEqual(InkComment.objects.count(), 0)
        self.post_parent_comment()
        self.assertEqual(InkComment.objects.count(), 1)
        data = {
            "name": "Bob",
            "email": "fulanito@detal.com",
            "followup": True,
            "reply_to": self.cm.id,  # This exceeds max thread level.
            "comment": "Es war einmal eine kleine...",
            "honeypot": "",
        }
        data.update(self.form.initial)
        response = post_comment(data)
        self.assertEqual(InkComment.objects.count(), 1)  # Comment not added.
        self.assertEqual(response.status_code, 400)


_ink_model = "django_comments_ink.tests.models.MyComment"


class CommentCountTestCase(APITestCase):
    def setUp(self):
        self.article = Article.objects.create(
            title="September", slug="september", body="During September..."
        )

    def _send_request(self):
        kwargs = {"content_type": "tests-article", "object_pk": "1"}
        req = factory.get(reverse("comments-ink-api-count", kwargs=kwargs))
        view = CommentCount.as_view()
        return view(req, **kwargs)

    def test_get_count_shall_be_0(self):
        resp = self._send_request()
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.rendered_content, b'{"count":0}')

    def test_get_count_shall_be_2(self):
        thread_test_step_1(self.article)
        resp = self._send_request()
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.rendered_content, b'{"count":2}')

    @patch.multiple(
        "django_comments_ink.conf.settings", COMMENTS_INK_MODEL=_ink_model
    )
    def test_get_count_for_custom_comment_model_shall_be_2(self):
        thread_test_step_1(
            self.article, model=MyComment, title="Can't be empty 1"
        )
        resp = self._send_request()
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.rendered_content, b'{"count":2}')

    @patch.multiple("django_comments_ink.conf.settings", SITE_ID=2)
    def test_get_count_for_comments_sent_to_different_site(self):
        site2 = Site.objects.create(domain="site2.com", name="site2.com")
        thread_test_step_1(self.article)
        thread_test_step_1(self.article, site=site2)
        resp = self._send_request()
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.rendered_content, b'{"count":2}')

    @patch.multiple(
        "django_comments_ink.conf.settings",
        COMMENTS_HIDE_REMOVED=True,
        COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED=True,
    )
    def test_get_count_for_HIDE_REMOVED_case_1(self):
        # To find out what are the cases 1, 2 and 3, read the docs settings
        # page, section COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED.
        thread_test_step_1(self.article)
        thread_test_step_2(self.article)
        # Previous two lines create the following comments:
        #  content ->    cmt.id  thread_id  parent_id  level  order
        #   cm1,   ->     1         1          1        0      1
        #   cm3,   ->     3         1          1        1      2
        #   cm4,   ->     4         1          1        1      3
        #   cm2,   ->     2         2          2        0      1
        cm1 = InkComment.objects.get(pk=1)
        cm1.is_removed = True
        cm1.save()
        # After removing the cm1, both cm3 and cm4 have is_public=False.
        # Therefore the count should return 1 -> cm2. cm1 is hidden.
        resp = self._send_request()
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.rendered_content, b'{"count":1}')

    @patch.multiple(
        "django_comments_ink.conf.settings",
        COMMENTS_HIDE_REMOVED=False,
        COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED=True,
    )
    def test_get_count_for_HIDE_REMOVED_case_2(self):
        thread_test_step_1(self.article)
        thread_test_step_2(self.article)
        # These two lines create the following comments:
        # (  # content ->    cmt.id  thread_id  parent_id  level  order
        #     cm1,   # ->     1         1          1        0      1
        #     cm3,   # ->     3         1          1        1      2
        #     cm4,   # ->     4         1          1        1      3
        #     cm2,   # ->     2         2          2        0      1
        # ) = InkComment.objects.all()
        #
        cm1 = InkComment.objects.get(pk=1)
        cm1.is_removed = True
        cm1.save()
        # After removing the cm1, both cm3 and cm4 have is_public=False,
        # as COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED is True. Therefore the
        # count should return 2 -> (cm1, cm2). cm1 is not hidden due to
        # COMMENTS_HIDE_REMOVED being False (a message will be displayed
        # saying that the comment has been removed, but the message won't be
        # removed from the queryset).
        resp = self._send_request()
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.rendered_content, b'{"count":2}')

    @patch.multiple(
        "django_comments_ink.conf.settings",
        COMMENTS_HIDE_REMOVED=False,
        COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED=False,
    )
    def test_get_count_for_HIDE_REMOVED_case_3(self):
        model_app_label = get_model()._meta.label
        # The function publish_or_withhold_on_pre_save is only called if
        # COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED are True.
        pre_save.disconnect(
            publish_or_withhold_on_pre_save, sender=model_app_label
        )
        thread_test_step_1(self.article)
        thread_test_step_2(self.article)
        #
        # These two lines create the following comments:
        #
        # (  # content ->    cmt.id  thread_id  parent_id  level  order
        #     cm1,   # ->     1         1          1        0      1
        #     cm3,   # ->     3         1          1        1      2
        #     cm4,   # ->     4         1          1        1      3
        #     cm2,   # ->     2         2          2        0      1
        # ) = InkComment.objects.all()

        cm1 = InkComment.objects.get(pk=1)
        cm1.is_removed = True
        cm1.save()
        # After removing the cm1, both cm3 and cm4 must remain visible,
        # as COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED is False.
        resp = self._send_request()
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.rendered_content, b'{"count":4}')

        # Re-connect the function for the following tests.
        pre_save.connect(
            publish_or_withhold_on_pre_save, sender=model_app_label
        )


class CommentListTestCase(APITestCase):
    def setUp(self):
        self.article = Article.objects.create(
            title="September", slug="september", body="During September..."
        )

    def _send_request(self):
        kwargs = {"content_type": "tests-article", "object_pk": "1"}
        req = factory.get(reverse("comments-ink-api-list", kwargs=kwargs))
        view = CommentList.as_view()
        return view(req, **kwargs)

    def test_get_list(self):
        thread_test_step_1(self.article)  # Sends 2 comments.
        thread_test_step_2(self.article)  # Sends 2 comments.
        thread_test_step_3(self.article)  # Sends 1 comment.
        # -> content:   cmt.id  thread_id  parent_id  level  order
        # cm1,   # ->      1         1          1        0      1
        # cm3,   # ->      3         1          1        1      2
        # cm4,   # ->      4         1          1        1      3
        # cm2,   # ->      2         2          2        0      1
        # cm5    # ->      5         2          2        1      2
        resp = self._send_request()
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.rendered_content)
        self.assertEqual(len(data), 5)
        for cm, cm_id in zip(data, [1, 3, 4, 2, 5]):
            self.assertEqual(cm["id"], cm_id)

    # Missing enhancement: extend the capacity to customize the comment model
    # to the API, so that a comment model can be used in combination with its
    # own serializer, and test it here.

    @patch.multiple("django_comments_ink.conf.settings", SITE_ID=2)
    def test_get_list_for_a_second_site(self):
        site2 = Site.objects.create(domain="site2.com", name="site2.com")

        # Send nested comments to the article in the site1.
        thread_test_step_1(self.article)
        thread_test_step_2(self.article)
        thread_test_step_3(self.article)

        # Send some nested comments to the article in the site2.
        thread_test_step_1(self.article, site=site2)

        resp = self._send_request()
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.rendered_content)
        self.assertEqual(len(data), 2)
        for cm, cm_id in zip(data, [6, 7]):
            self.assertEqual(cm["id"], cm_id)

    @patch.multiple(
        "django_comments_ink.conf.settings",
        COMMENTS_HIDE_REMOVED=True,
        COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED=True,
    )
    def test_get_list_for_HIDE_REMOVED_case_1(self):
        thread_test_step_1(self.article)
        thread_test_step_2(self.article)
        # Previous two lines create the following comments:
        #  content ->    cmt.id  thread_id  parent_id  level  order
        #   cm1,   ->     1         1          1        0      1
        #   cm3,   ->     3         1          1        1      2
        #   cm4,   ->     4         1          1        1      3
        #   cm2,   ->     2         2          2        0      1
        cm1 = InkComment.objects.get(pk=1)
        cm1.is_removed = True
        cm1.save()
        # After removing cm1, both cm3 and cm4 have is_public=False.
        # Therefore the list of comments should contain only cm2.
        resp = self._send_request()
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.rendered_content)
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]["id"], 2)

    @patch.multiple(
        "django_comments_ink.conf.settings",
        COMMENTS_HIDE_REMOVED=False,
        COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED=True,
    )
    def test_get_list_for_HIDE_REMOVED_case_2(self):
        thread_test_step_1(self.article)
        thread_test_step_2(self.article)
        # These two lines create the following comments:
        # (  # content ->    cmt.id  thread_id  parent_id  level  order
        #     cm1,   # ->     1         1          1        0      1
        #     cm3,   # ->     3         1          1        1      2
        #     cm4,   # ->     4         1          1        1      3
        #     cm2,   # ->     2         2          2        0      1
        # ) = InkComment.objects.all()
        #
        cm1 = InkComment.objects.get(pk=1)
        cm1.is_removed = True
        cm1.save()
        # After removing the cm1, both cm3 and cm4 have is_public=False,
        # as COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED is True. Therefore the
        # count should return 2 -> (cm1, cm2). cm1 is not hidden due to
        # COMMENTS_HIDE_REMOVED being False (a message will be displayed
        # saying that the comment has been removed, but the message won't be
        # removed from the queryset).
        resp = self._send_request()
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.rendered_content)
        self.assertEqual(len(data), 2)
        for cm, cm_id in zip(data, [1, 2]):
            self.assertEqual(cm["id"], cm_id)

    @patch.multiple(
        "django_comments_ink.conf.settings",
        COMMENTS_HIDE_REMOVED=False,
        COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED=False,
    )
    def test_get_list_for_HIDE_REMOVED_case_3(self):
        model_app_label = get_model()._meta.label
        # The function publish_or_withhold_on_pre_save is only called if
        # COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED are True.
        pre_save.disconnect(
            publish_or_withhold_on_pre_save, sender=model_app_label
        )
        thread_test_step_1(self.article)
        thread_test_step_2(self.article)
        # These two lines create the following comments:
        # (  # content ->    cmt.id  thread_id  parent_id  level  order
        #     cm1,   # ->     1         1          1        0      1
        #     cm3,   # ->     3         1          1        1      2
        #     cm4,   # ->     4         1          1        1      3
        #     cm2,   # ->     2         2          2        0      1
        # ) = InkComment.objects.all()
        cm1 = InkComment.objects.get(pk=1)
        cm1.is_removed = True
        cm1.save()
        # After removing the cm1, both cm3 and cm4 remain visible,
        # as COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED is False.
        resp = self._send_request()
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.rendered_content)
        self.assertEqual(len(data), 4)
        for cm, cm_id in zip(data, [1, 3, 4, 2]):
            self.assertEqual(cm["id"], cm_id)
        # Re-connect the function for the following tests.
        pre_save.connect(
            publish_or_withhold_on_pre_save, sender=model_app_label
        )


@pytest.mark.django_db
def test_CommentList_handles_no_ContentType():
    kwargs = {"content_type": "this-that", "object_pk": "1"}
    request = factory.get(reverse("comments-ink-api-list", kwargs=kwargs))
    view = CommentList.as_view()
    response = view(request, **kwargs)
    assert response.status_code == 200
    assert response.rendered_content == b"[]"


@pytest.mark.django_db
def test_CommentCount_settings_has_no_SITE_ID(monkeypatch):
    # Remove 'SITE_ID' from settings module, to test that method 'get_queryset'
    # in CommentCount class uses function 'utils.get_current_site_id' to get
    # the site_id.
    monkeypatch.delattr(settings, name="SITE_ID")
    kwargs = {"content_type": "tests-article", "object_pk": "1"}
    request = factory.get(reverse("comments-ink-api-count", kwargs=kwargs))
    view = CommentCount.as_view()
    response = view(request, **kwargs)
    assert response.status_code == 200
    assert response.rendered_content == b'{"count":0}'



from unittest.mock import patch

from django.test import TestCase


class GetVersionTestCase(TestCase):
    @patch("django_comments_ink.VERSION", (2, 8, 0, "f", 0))
    def test_get_version_when_patch_equal_to_zero(self):
        from django_comments_ink import get_version

        self.assertEqual(get_version(), "2.8.0")

    @patch("django_comments_ink.VERSION", (2, 8, 1, "f", 0))
    def test_get_version_when_patch_greater_than_zero(self):
        from django_comments_ink import get_version

        self.assertEqual(get_version(), "2.8.1")

    @patch("django_comments_ink.VERSION", (2, 8, 1, 9, 8))
    def test_get_version_when_version_three_not_f(self):
        from django_comments_ink import get_version

        self.assertEqual(get_version(), "2.8.198")



from datetime import datetime

from django.db import models
from django.urls import reverse
from django_comments_ink.conf import settings
from django_comments_ink.models import InkComment
from django_comments_ink.moderation import InkCommentModerator, moderator


class PublicManager(models.Manager):
    """Returns published articles that are not in the future."""

    def published(self):
        return self.get_query_set().filter(publish__lte=datetime.now())


class Article(models.Model):
    """Article, that accepts comments."""

    title = models.CharField("title", max_length=200)
    slug = models.SlugField("slug", unique_for_date="publish")
    body = models.TextField("body")
    allow_comments = models.BooleanField("allow comments", default=True)
    publish = models.DateTimeField("publish", default=datetime.now)

    objects = PublicManager()

    class Meta:
        db_table = "demo_articles"
        ordering = ("-publish",)

    def get_absolute_url(self):
        return reverse(
            "article-detail",
            kwargs={
                "year": self.publish.year,
                "month": int(self.publish.strftime("%m").lower()),
                "day": self.publish.day,
                "slug": self.slug,
            },
        )


class Diary(models.Model):
    """Diary, that accepts comments."""

    body = models.TextField("body")
    allow_comments = models.BooleanField("allow comments", default=True)
    publish = models.DateTimeField("publish", default=datetime.now)

    objects = PublicManager()

    class Meta:
        db_table = "demo_diary"
        ordering = ("-publish",)

    def get_absolute_url(self):
        return reverse(
            "diary-detail",
            kwargs={
                "year": self.publish.year,
                "month": int(self.publish.strftime("%m").lower()),
                "day": self.publish.day,
            },
        )


class DiaryCommentModerator(InkCommentModerator):
    email_notification = True
    enable_field = "allow_comments"
    auto_moderate_field = "publish"
    moderate_after = 2
    removal_suggestion_notification = True


moderator.register(Diary, DiaryCommentModerator)


def authorize_api_post_comment(sender, comment, request, **kwargs):
    if (request.user and request.user.is_authenticated) or (
        request.auth and request.auth == settings.MY_DRF_AUTH_TOKEN
    ):
        return True
    else:
        return False


def comment_will_be_rejected(sender, comment, request, **kwargs):
    return False


class MyComment(InkComment):
    title = models.CharField(max_length=256)



from django.urls import path, re_path

from .views import (
    CommentCount,
    CommentCreate,
    CommentList,
    CommentReactionAuthorList,
    CreateReportFlag,
    ObjectReactionAuthorList,
    PostCommentReaction,
)

urlpatterns = [
    path("comment/", CommentCreate.as_view(), name="comments-ink-api-create"),
    path(
        "react/",
        PostCommentReaction.as_view(),
        name="comments-ink-api-react",
    ),
    path("flag/", CreateReportFlag.as_view(), name="comments-ink-api-flag"),
    re_path(
        r"^(?P<comment_pk>[\d]+)/(?P<reaction_value>[\w\+\-]+)/$",
        CommentReactionAuthorList.as_view(),
        name="comments-ink-comment-reaction-authors",
    ),
    # -----------------------------------------------------------------
    # The following 3 re_path entries read the content type as
    # <applabel>-<model>, and the object ID to which comments
    # have been sent.
    # List the comments sent to the <ctype>/<object_pk>.
    re_path(
        r"^(?P<content_type>\w+[-]{1}\w+)/(?P<object_pk>[-\w]+)/$",
        CommentList.as_view(),
        name="comments-ink-api-list",
    ),
    # Number of comments sent to the <ctype>/<object_pk>.
    re_path(
        r"^(?P<content_type>\w+[-]{1}\w+)/(?P<object_pk>[-\w]+)/count/$",
        CommentCount.as_view(),
        name="comments-ink-api-count",
    ),
    # List users that reacted with <reaction_value> to the post
    # identified with <ctype>/<object_pk>.
    re_path(
        r"^(?P<content_type>\w+[-]{1}\w+)/"
        r"(?P<object_pk>[-\w]+)/"
        r"(?P<reaction_value>[\w\+\-]+)/$",
        ObjectReactionAuthorList.as_view(),
        name="comments-ink-api-object-reaction-authors",
    ),
]



# Generated by Django 3.2.13 on 2022-04-25 11:45

import django.db.models.deletion
import django_comments_ink.models
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
        ("django_comments", "0004_add_object_pk_is_removed_index"),
        ("sites", "0002_alter_domain_unique"),
    ]

    operations = [
        migrations.CreateModel(
            name="BlackListedDomain",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("domain", models.CharField(db_index=True, max_length=200)),
            ],
            options={
                "ordering": ("domain",),
            },
        ),
        migrations.CreateModel(
            name="CommentReaction",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "reaction",
                    django_comments_ink.models.ReactionField(
                        db_index=True, verbose_name="reaction"
                    ),
                ),
                ("counter", models.IntegerField(default=0)),
            ],
            options={
                "verbose_name": "comment reactions",
                "verbose_name_plural": "comments reactions",
            },
        ),
        migrations.CreateModel(
            name="InkComment",
            fields=[
                (
                    "comment_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_comments.comment",
                    ),
                ),
                ("thread_id", models.IntegerField(db_index=True, default=0)),
                ("parent_id", models.IntegerField(default=0)),
                ("level", models.SmallIntegerField(default=0)),
                ("order", models.IntegerField(db_index=True, default=1)),
                (
                    "followup",
                    models.BooleanField(
                        blank=True,
                        default=False,
                        help_text="Notify follow-up comments",
                    ),
                ),
                ("nested_count", models.IntegerField(db_index=True, default=0)),
            ],
            options={
                "verbose_name": "comment",
                "verbose_name_plural": "comments",
                "ordering": ("submit_date",),
                "permissions": [("can_moderate", "Can moderate comments")],
                "abstract": False,
            },
            bases=("django_comments.comment",),
        ),
        migrations.CreateModel(
            name="ObjectReaction",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "reaction",
                    django_comments_ink.models.ReactionField(
                        db_index=True, verbose_name="reaction"
                    ),
                ),
                (
                    "object_pk",
                    models.CharField(
                        db_index=True, max_length=64, verbose_name="object ID"
                    ),
                ),
                ("counter", models.IntegerField(default=0)),
            ],
            options={
                "verbose_name": "object reactions",
                "verbose_name_plural": "objects reactions",
            },
        ),
        migrations.CreateModel(
            name="ObjectReactionAuthor",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "author",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "reaction",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_comments_ink.objectreaction",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="objectreaction",
            name="authors",
            field=models.ManyToManyField(
                through="django_comments_ink.ObjectReactionAuthor",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="objectreaction",
            name="content_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="content_type_set_for_objectreaction",
                to="contenttypes.contenttype",
                verbose_name="content type",
            ),
        ),
        migrations.AddField(
            model_name="objectreaction",
            name="site",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="sites.site"
            ),
        ),
        migrations.CreateModel(
            name="CommentReactionAuthor",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "author",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "reaction",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_comments_ink.commentreaction",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="commentreaction",
            name="authors",
            field=models.ManyToManyField(
                through="django_comments_ink.CommentReactionAuthor",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="commentreaction",
            name="comment",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="reactions",
                to="django_comments_ink.inkcomment",
                verbose_name="reactions",
            ),
        ),
    ]



"""
Signals relating to django-comments-ink.
"""
from django.dispatch import Signal

# Sent just after a comment has been verified.
confirmation_received = Signal()


# Sent just after a user has muted a comments thread.
comment_thread_muted = Signal()


# Sent before the data in the REST POST comment form is validated.
# A receiver returning True will suffice to automatically add valid values
# to the CommentSecurityForm fields 'timestamp' and 'security_hash'. The
# intention is to combine a receiver with a django-rest-framework
# authentication class, and return True when the request.auth is not None.
should_request_be_authorized = Signal()


# Sent after a comment got a reaction.
comment_got_a_reaction = Signal()

# Sent after a comment got a reaction.
object_got_a_reaction = Signal()



import collections
from datetime import datetime

import pytest
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django_comments_ink import get_model
from django_comments_ink.paginator import CommentsPaginator

InkComment = get_model()


def test_paginator_only_accepts_queryset():
    with pytest.raises(TypeError):
        CommentsPaginator([1, 2, 3], 10, 3)


@pytest.mark.django_db
def test_paginator_example_1(an_article):
    """Tests the Example 1 (detailed in the paginator.py __docs__)."""
    article_ct = ContentType.objects.get(app_label="tests", model="article")
    site = Site.objects.get(pk=1)
    attrs = {
        "content_type": article_ct,
        "object_pk": an_article.pk,
        "content_object": an_article,
        "site": site,
        "comment": f"another comment to article {an_article.pk}",
        "submit_date": datetime.now(),
    }

    # Create 8 comments at level 0.
    for cm_level_0 in range(8):
        InkComment.objects.create(**attrs)

    # Verify that the 8 comments are of level 0.
    cmts_level_0 = InkComment.objects.all()
    for index, cm in enumerate(cmts_level_0):
        assert cm.pk == index + 1
        assert cm.level == 0
        assert cm.parent_id == cm.thread_id == cm.pk

    # Add the following number of child comments to the previous cmts_level_0.
    children_number = [10, 10, 10, 10, 10, 10, 5, 4]
    for index, cmt_level_0 in enumerate(cmts_level_0):
        for child in range(children_number[index]):
            InkComment.objects.create(**attrs, parent_id=cmt_level_0.pk)

    # Verify that the children comments are nested into the cmts_level_0.
    cmts_level_0 = InkComment.objects.filter(level=0)
    for index, cmt_level_0 in enumerate(cmts_level_0):
        assert cmt_level_0.pk == index + 1
        assert cmt_level_0.nested_count == children_number[index]

    # Test output from paginator.
    queryset = InkComment.objects.all()
    page_size = 25
    orphans = 10
    paginator = CommentsPaginator(queryset, page_size, orphans)
    assert paginator.count == queryset.count()

    # Page 1 shall contain 22 comments.
    page = paginator.page(1)
    assert page.object_list.count() == 22
    comment_list = [cm.parent_id for cm in page.object_list]
    assert set(comment_list) == set([1, 2])
    counter = collections.Counter(comment_list)
    assert counter[1] == 11
    assert counter[2] == 11

    # Page 2 shall contain 22 comments.
    page = paginator.page(2)
    assert page.object_list.count() == 22
    comment_list = [cm.parent_id for cm in page.object_list]
    assert set(comment_list) == set([3, 4])
    counter = collections.Counter(comment_list)
    assert counter[3] == 11
    assert counter[4] == 11

    # Page 3 shall contain 22 comments.
    page = paginator.page(3)
    assert page.object_list.count() == 33
    comment_list = [cm.parent_id for cm in page.object_list]
    assert set(comment_list) == set([5, 6, 7, 8])
    counter = collections.Counter(comment_list)
    assert counter[5] == 11
    assert counter[6] == 11
    assert counter[7] == 6
    assert counter[8] == 5


@pytest.mark.django_db
def test_paginator_example_2(an_article):
    """Tests the Example 2 (detailed in the paginator.py __docs__)."""
    article_ct = ContentType.objects.get(app_label="tests", model="article")
    site = Site.objects.get(pk=1)
    attrs = {
        "content_type": article_ct,
        "object_pk": an_article.pk,
        "content_object": an_article,
        "site": site,
        "comment": f"another comment to article {an_article.pk}",
        "submit_date": datetime.now(),
    }

    # Create 3 comments at level 0.
    for cm_level_0 in range(3):
        InkComment.objects.create(**attrs)

    # Verify that the 3 comments are of level 0.
    cmts_level_0 = InkComment.objects.all()
    for index, cm in enumerate(cmts_level_0):
        assert cm.pk == index + 1
        assert cm.level == 0
        assert cm.parent_id == cm.thread_id == cm.pk

    # Add the following number of child comments to the previous cmts_level_0.
    children_number = [24, 24, 8]
    for index, cmt_level_0 in enumerate(cmts_level_0):
        for child in range(children_number[index]):
            InkComment.objects.create(**attrs, parent_id=cmt_level_0.pk)

    # Verify that the children comments are nested into the cmts_level_0.
    cmts_level_0 = InkComment.objects.filter(level=0)
    for index, cmt_level_0 in enumerate(cmts_level_0):
        assert cmt_level_0.pk == index + 1
        assert cmt_level_0.nested_count == children_number[index]

    # Test output from paginator.
    queryset = InkComment.objects.all()
    page_size = 25
    orphans = 10
    paginator = CommentsPaginator(queryset, page_size, orphans)
    assert paginator.count == queryset.count()

    # Page 1 shall contain 25 comments.
    page = paginator.page(1)
    assert page.object_list.count() == 25
    comment_list = [cm.parent_id for cm in page.object_list]
    assert set(comment_list) == set([1])
    counter = collections.Counter(comment_list)
    assert counter[1] == 25

    # Page 2 shall contain 22 comments.
    page = paginator.page(2)
    assert page.object_list.count() == 34
    comment_list = [cm.parent_id for cm in page.object_list]
    assert set(comment_list) == set([2, 3])
    counter = collections.Counter(comment_list)
    assert counter[2] == 25
    assert counter[3] == 9


@pytest.mark.django_db
def test_paginator_example_3(an_article):
    """Tests the Example 3 (detailed in the paginator.py __docs__)."""
    article_ct = ContentType.objects.get(app_label="tests", model="article")
    site = Site.objects.get(pk=1)
    attrs = {
        "content_type": article_ct,
        "object_pk": an_article.pk,
        "content_object": an_article,
        "site": site,
        "comment": f"another comment to article {an_article.pk}",
        "submit_date": datetime.now(),
    }

    # Create 3 comments at level 0.
    for cm_level_0 in range(3):
        InkComment.objects.create(**attrs)

    # Verify that the 3 comments are of level 0.
    cmts_level_0 = InkComment.objects.all()
    for index, cm in enumerate(cmts_level_0):
        assert cm.pk == index + 1
        assert cm.level == 0
        assert cm.parent_id == cm.thread_id == cm.pk

    # Add the following number of child comments to the previous cmts_level_0.
    children_number = [24, 8, 8]
    for index, cmt_level_0 in enumerate(cmts_level_0):
        for child in range(children_number[index]):
            InkComment.objects.create(**attrs, parent_id=cmt_level_0.pk)

    # Verify that the children comments are nested into the cmts_level_0.
    cmts_level_0 = InkComment.objects.filter(level=0)
    for index, cmt_level_0 in enumerate(cmts_level_0):
        assert cmt_level_0.pk == index + 1
        assert cmt_level_0.nested_count == children_number[index]

    # Test output from paginator.
    queryset = InkComment.objects.all()
    page_size = 25
    orphans = 10
    paginator = CommentsPaginator(queryset, page_size, orphans)
    assert paginator.count == queryset.count()

    # Page 1 shall contain 25 comments.
    page = paginator.page(1)
    assert page.object_list.count() == 25
    comment_list = [cm.parent_id for cm in page.object_list]
    assert set(comment_list) == set([1])
    counter = collections.Counter(comment_list)
    assert counter[1] == 25

    # Page 2 shall contain 22 comments.
    page = paginator.page(2)
    assert page.object_list.count() == 18
    comment_list = [cm.parent_id for cm in page.object_list]
    assert set(comment_list) == set([2, 3])
    counter = collections.Counter(comment_list)
    assert counter[2] == 9
    assert counter[3] == 9


@pytest.mark.django_db
def test_paginator_allow_empty_first_page():
    qs = InkComment.objects.all()
    paginator = CommentsPaginator(
        qs, 10, orphans=3, allow_empty_first_page=False
    )
    assert paginator.num_pages == 0



import json

import pytest
from django.contrib.auth.models import AnonymousUser, Permission
from django.contrib.contenttypes.models import ContentType
from django.urls import reverse
from django_comments.models import Comment
from django_comments_ink import utils
from django_comments_ink.api import frontend
from django_comments_ink.conf import settings
from django_comments_ink.models import max_thread_level_for_content_type


@pytest.mark.django_db
def test_comment_box_props_with_object_and_user(an_user, an_article):
    """
    Verify the default props obtained for an_article without comments.
    """
    props = frontend.comments_api_props(an_article, an_user)

    ctype = ContentType.objects.get_for_model(an_article)
    ctype_slug = "%s-%s" % (ctype.app_label, ctype.model)

    # Verify that the props dictionary contains a form dictionary
    # with the same keys as the CommentSecurityForm.
    assert "content_type" in props["form"]
    assert "object_pk" in props["form"]
    assert "timestamp" in props["form"]
    assert "security_hash" in props["form"]

    assert props["comment_count"] == 0
    assert props["input_allowed"] == True
    assert props["current_user"] == "1:joe"
    assert props["request_name"] == False  # an_user.get_full_name() != "".
    assert props["request_email_address"] == False  # an_user.email != None.
    assert props["is_authenticated"] == True
    assert props["who_can_post"] == "all"
    assert props["comment_flagging_enabled"] == False
    assert props["comment_reactions_enabled"] == False
    assert props["object_reactions_enabled"] == False
    assert props["can_moderate"] == False
    assert props["react_url"] == reverse("comments-ink-api-react")
    assert props["delete_url"] == reverse("comments-delete", args=(0,))
    assert props["reply_url"] == reverse(
        "comments-ink-reply", kwargs={"cid": 0}
    )
    assert props["flag_url"] == reverse("comments-ink-api-flag")
    assert props["list_url"] == reverse(
        "comments-ink-api-list",
        kwargs={"content_type": ctype_slug, "object_pk": an_article.id},
    )
    assert props["count_url"] == reverse(
        "comments-ink-api-count",
        kwargs={"content_type": ctype_slug, "object_pk": an_article.id},
    )
    assert props["send_url"] == reverse("comments-ink-api-create")
    assert props["default_followup"] == settings.COMMENTS_INK_DEFAULT_FOLLOWUP
    assert props["html_id_suffix"] == utils.get_html_id_suffix(an_article)
    assert props["max_thread_level"] == max_thread_level_for_content_type(ctype)


def check_input_allowed(object):
    return False


@pytest.mark.django_db
def test_comment_box_props_with_new_app_model_options(
    an_user, an_article, monkeypatch
):
    """
    Modify setting COMMENTS_INK_APP_MODEL_OPTIONS to contain an entry for
    'check_input_allowed' pointing to a function that will return False. The
    dictionary returned by comments_api_props must have a key 'input_allowed' with a False value. Also modify the rest of the values provided by the
    setting to check that comments_api_props reads them correctly.
    """
    check_f = "django_comments_ink.tests.test_frontend.check_input_allowed"
    app_model_options = {
        "who_can_post": "users",
        "check_input_allowed": check_f,
        "comment_flagging_enabled": True,
        "comment_reactions_enabled": True,
        "object_reactions_enabled": True,
    }
    monkeypatch.setattr(
        frontend, "get_app_model_options", lambda **kwargs: app_model_options
    )
    props = frontend.comments_api_props(an_article, an_user)
    assert props["input_allowed"] == False
    assert props["who_can_post"] == "users"
    assert props["comment_flagging_enabled"] == True
    assert props["comment_reactions_enabled"] == True
    assert props["object_reactions_enabled"] == True


@pytest.mark.django_db
def test_comment_box_props_with_new_max_thread_level(
    an_user, an_article, monkeypatch
):
    monkeypatch.setattr(
        frontend, "max_thread_level_for_content_type", lambda *args: 1
    )
    props = frontend.comments_api_props(an_article, an_user)
    assert props["max_thread_level"] == 1


@pytest.mark.django_db
def test_comment_box_props_with_anonymous_user(an_article):
    anonymous_user = AnonymousUser()
    props = frontend.comments_api_props(an_article, anonymous_user)
    assert props["current_user"] == "0:Anonymous"
    assert props["is_authenticated"] == False
    assert "login_url" in props
    assert props["login_url"] == settings.LOGIN_URL


@pytest.mark.django_db
def test_comment_box_props_with_user_that_can_moderate(an_user, an_article):
    ct_comments = ContentType.objects.get_for_model(Comment)
    can_moderate = Permission.objects.get(
        codename="can_moderate", content_type=ct_comments
    )
    an_user.user_permissions.add(can_moderate)
    props = frontend.comments_api_props(an_article, an_user)
    assert props["can_moderate"] == True


@pytest.mark.django_db
def test_comment_box_props_response(an_user, an_article):
    from rest_framework.response import Response

    response = frontend.comments_api_props_response(
        an_article, an_user, request=None
    )
    props = frontend.comments_api_props(an_article, an_user)
    assert isinstance(response, Response)
    assert response.data == props



from django.contrib.contenttypes.models import ContentType
from django.utils.module_loading import import_string
from django_comments.forms import CommentSecurityForm
from django_comments_ink import get_model as get_comment_model
from django_comments_ink.conf import settings
from django_comments_ink.models import max_thread_level_for_content_type
from django_comments_ink.utils import (
    get_app_model_options,
    get_current_site_id,
    get_html_id_suffix,
)
from rest_framework.response import Response
from rest_framework.reverse import reverse

InkComment = get_comment_model()


def comments_api_props(obj, user, request=None):
    """
    Returns a JSON object with the initial props for the CommentBox component.

    The returned JSON object contains the following attributes::
        {
            comment_count: <int>,  Count of comments posted to the object.
            input_allowed: <bool>, Whether to allow comments to this post.
            current_user: <str as "user_id:user_name">,
            is_authenticated: <bool>,  Whether current_user is authenticated.
            request_name: <bool>,  True when auth user has no actual name.
            request_email_address: <bool>,  True when auth user has no email.
            is_authenticated: <bool>,
            who_can_post: 'users' | 'all',
            comment_flagging_enabled: <bool>,
            comment_reactions_enabled: <bool>,
            object_reactions_enabled: <bool>,
            can_moderate: <bool>,  Whether current_user can moderate.
            react_url: <api-url-to-send-reactions-to-comments>,
            delete_url: <api-url-for-moderators-to-remove-comment>,
            reply_url: <api-url-to-reply-comments>,
            flag_url: <api-url-to-suggest-comment-removal>,
            list_url: <api-url-to-list-comments>,
            count_url: <api-url-to-count-comments>,
            send_url: <api-url-to-send-a-comment>,
            form: {
                content_type: <value>,
                object_pk: <value>,
                timestamp: <value>,
                security_hash: <value>
            },
            login_url: <only_when_user_is_not_authenticated>,
            html_id_suffix: <html_element_id_suffix>,
            max_thread_level: max_thread_level for the content type of the obj.
            comments_page_qs_param: <string>, name of comment's page qs param.
        }
    """
    form = CommentSecurityForm(obj)
    ctype = ContentType.objects.get_for_model(obj)
    queryset = InkComment.objects.filter(
        content_type=ctype,
        object_pk=obj.pk,
        site__pk=get_current_site_id(request),
        is_public=True,
    )
    ctype_slug = "%s-%s" % (ctype.app_label, ctype.model)
    app_model = "%s.%s" % (ctype.app_label, ctype.model)
    options = get_app_model_options(content_type=app_model)
    check_input_allowed_str = options.pop("check_input_allowed")
    check_func = import_string(check_input_allowed_str)
    d = {
        "comment_count": queryset.count(),
        "input_allowed": check_func(obj),
        "current_user": "0:Anonymous",
        "request_name": False,
        "request_email_address": False,
        "is_authenticated": False,
        "who_can_post": options["who_can_post"],
        "comment_flagging_enabled": options["comment_flagging_enabled"],
        "comment_reactions_enabled": options["comment_reactions_enabled"],
        "object_reactions_enabled": options["object_reactions_enabled"],
        "can_moderate": False,
        "react_url": reverse("comments-ink-api-react"),
        "delete_url": reverse("comments-delete", args=(0,)),
        "reply_url": reverse("comments-ink-reply", kwargs={"cid": 0}),
        "flag_url": reverse("comments-ink-api-flag"),
        "list_url": reverse(
            "comments-ink-api-list",
            kwargs={"content_type": ctype_slug, "object_pk": obj.id},
        ),
        "count_url": reverse(
            "comments-ink-api-count",
            kwargs={"content_type": ctype_slug, "object_pk": obj.id},
        ),
        "send_url": reverse("comments-ink-api-create"),
        "form": {
            "content_type": form["content_type"].value(),
            "object_pk": form["object_pk"].value(),
            "timestamp": form["timestamp"].value(),
            "security_hash": form["security_hash"].value(),
        },
        "default_followup": settings.COMMENTS_INK_DEFAULT_FOLLOWUP,
        "html_id_suffix": get_html_id_suffix(obj),
        "max_thread_level": max_thread_level_for_content_type(ctype),
        "comments_page_qs_param": settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM,
    }
    if user and user.is_authenticated:
        d["current_user"] = "%d:%s" % (
            user.pk,
            settings.COMMENTS_INK_API_USER_REPR(user),
        )
        d["is_authenticated"] = True
        d["can_moderate"] = user.has_perm("django_comments.can_moderate")
        d["request_name"] = True if not len(user.get_full_name()) else False
        d["request_email_address"] = True if not user.email else False
    else:
        d["login_url"] = settings.LOGIN_URL

    return d


def comments_api_props_response(obj, user, request):
    """Return a Response containing React props for use with client-side JS.
    Can add as an extra action to a ViewSet as follows:

        @action(detail=True, methods=['get'],
                permission_classes=[permissions.IsAuthenticated])
        def comments_api_props(self, request, *args, **kwargs):
            return comments_api_props_response(self.get_object(),
                                               request.user, request)
    """
    return Response(data=comments_api_props(obj, user, request=request))



from django.core.cache.backends.base import BaseCache
from django_comments_ink import caching


def test_get_cache_returns_a_cache(monkeypatch):
    monkeypatch.setattr(caching, "dci_cache", None)
    cache = caching.get_cache()
    assert cache != None
    assert isinstance(cache, BaseCache)


def test_get_cache_can_be_none(monkeypatch):
    # I'm not sure if this can happen, but I test it just in case.
    # I mean, that accesing django's caches I can get a None
    # instead of an instance of a backend cache.
    my_cache = {"dci": None}
    monkeypatch.setattr(caching, "dci_cache", None)
    monkeypatch.setattr(caching, "caches", my_cache)
    cache = caching.get_cache()
    assert cache == None


def test_clear_cache_returns_False(monkeypatch):
    my_cache = {"dci": None}
    monkeypatch.setattr(caching, "dci_cache", None)
    monkeypatch.setattr(caching, "caches", my_cache)
    assert caching.clear_cache(1, 2, 3) == False


def test_clear_cache_returns_True(monkeypatch):
    monkeypatch.setattr(caching, "dci_cache", None)
    assert caching.clear_cache(1, 2, 3) == True



from __future__ import unicode_literals

import importlib
import re
from datetime import datetime, timedelta
from unittest.mock import patch

import django_comments
from django.contrib.auth.models import AnonymousUser, Permission, User
from django.contrib.contenttypes.models import ContentType
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django_comments.models import Comment, CommentFlag
from django_comments.views.moderation import delete
from django_comments_ink import views
from django_comments_ink.models import InkComment
from django_comments_ink.tests.models import (
    Article,
    Diary,
    DiaryCommentModerator,
)
from django_comments_ink.tests.test_views import (
    confirm_comment_url,
    post_article_comment,
    post_diary_comment,
)

request_factory = RequestFactory()


send_mail = ""  # string to send_mail function to patch


if importlib.util.find_spec("django_comments"):
    send_mail = "django_comments.moderation.send_mail"
else:
    send_mail = "django.contrib.comments.moderation.send_mail"


class ModeratorApprovesComment(TestCase):
    def setUp(self):
        self.patcher_app1 = patch(send_mail)
        self.patcher_app2 = patch("django_comments_ink.views.utils.send_mail")
        self.mailer_app1 = self.patcher_app1.start()
        self.mailer_app2 = self.patcher_app2.start()
        self.diary_entry = Diary.objects.create(
            body="What I did on October...",
            allow_comments=True,
            publish=datetime.now(),
        )
        self.form = django_comments.get_form()(self.diary_entry)

    def tearDown(self):
        self.patcher_app1.stop()
        self.patcher_app2.stop()

    def post_valid_data(self, auth_user=None):
        data = {
            "name": "Bob",
            "email": "bob@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal eine kleine...",
        }
        data.update(self.form.initial)
        response = post_diary_comment(
            data, self.diary_entry, auth_user=auth_user
        )
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith("/comments/posted/?c="))

    def test_moderation_with_registered_user(self):
        user = User.objects.create_user("bob", "bob@example.com", "pwd")
        self.assertTrue(self.mailer_app1.call_count == 0)
        self.post_valid_data(user)
        # Moderation class:
        # django_comments_ink.tests.models.DiaryCommentModerator
        # must trigger an email once comment has passed moderation.
        self.assertTrue(self.mailer_app1.call_count == 1)
        comment = django_comments.get_model().objects.for_app_models(
            "tests.diary"
        )[0]
        self.assertTrue(comment.is_public is True)

    def test_moderation_with_unregistered_user(self):
        self.post_valid_data()
        self.assertTrue(self.mailer_app1.call_count == 0)
        self.assertTrue(self.mailer_app2.call_count == 1)
        mail_msg = self.mailer_app2.call_args[0][1]
        key = str(
            re.search(r"http://.+/confirm/(?P<key>[\S]+)/", mail_msg).group(
                "key"
            )
        )
        confirm_comment_url(key)
        self.assertTrue(self.mailer_app1.call_count == 1)
        self.assertTrue(self.mailer_app2.call_count == 1)
        comment = django_comments.get_model().objects.for_app_models(
            "tests.diary"
        )[0]
        self.assertTrue(comment.is_public is True)


class ModeratorHoldsComment(TestCase):
    def setUp(self):
        self.patcher_app1 = patch(send_mail)
        self.patcher_app2 = patch("django_comments_ink.views.utils.send_mail")
        self.mailer_app1 = self.patcher_app1.start()
        self.mailer_app2 = self.patcher_app2.start()
        self.diary_entry = Diary.objects.create(
            body="What I did Yesterday...",
            allow_comments=True,
            publish=datetime.now() - timedelta(days=5),
        )
        self.form = django_comments.get_form()(self.diary_entry)

    def tearDown(self):
        self.patcher_app1.stop()
        self.patcher_app2.stop()

    def post_valid_data(self, auth_user=None):
        data = {
            "name": "Bob",
            "email": "bob@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal eine kleine...",
        }
        data.update(self.form.initial)
        response = post_diary_comment(
            data, self.diary_entry, auth_user=auth_user
        )
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.startswith("/comments/posted/?c="))

    def test_moderation_with_registered_user(self):
        user = User.objects.create_user("bob", "bob@example.com", "pwd")
        self.assertTrue(self.mailer_app1.call_count == 0)
        self.post_valid_data(user)
        # Moderation class:
        # django_comments_ink.tests.models.DiaryCommentModerator
        # must trigger an email once comment has passed moderation.
        self.assertTrue(self.mailer_app1.call_count == 1)
        comment = django_comments.get_model().objects.for_app_models(
            "tests.diary"
        )[0]
        self.assertTrue(comment.is_public is False)

    def test_moderation_with_unregistered_user(self):
        self.post_valid_data()
        self.assertTrue(self.mailer_app1.call_count == 0)
        self.assertTrue(self.mailer_app2.call_count == 1)
        mail_msg = self.mailer_app2.call_args[0][1]
        key = str(
            re.search(r"http://.+/confirm/(?P<key>[\S]+)/", mail_msg).group(
                "key"
            )
        )
        confirm_comment_url(key)
        self.assertTrue(self.mailer_app1.call_count == 1)
        self.assertTrue(self.mailer_app2.call_count == 1)
        comment = django_comments.get_model().objects.for_app_models(
            "tests.diary"
        )[0]
        self.assertTrue(comment.is_public is False)


app_model_options_mock = {
    "tests.article": {
        "comment_flagging_enabled": True,
        "comment_reactions_enabled": False,
        "object_reactions_enabled": False,
    }
}


class FlaggingRemovalSuggestion(TestCase):
    """Scenario to test the flag removal_suggestion_notification"""

    def setUp(self):
        patcher = patch("django_comments_ink.moderation.send_mail")
        self.mailer = patcher.start()
        diary_entry = Diary.objects.create(
            body="What I did on October...",
            allow_comments=True,
            publish=datetime.now(),
        )
        self.user = User.objects.create_user("bob", "bob@example.com", "pwd")
        data = {
            "name": "Bob",
            "email": "bob@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmal eine kleine...",
        }
        form = django_comments.get_form()(diary_entry)
        data.update(form.initial)
        post_diary_comment(data, diary_entry, auth_user=self.user)
        # Also create an article and a comment to that article,
        # To test that flagging a comment of a model not registered
        # for moderation does not trigger the notify_removal_suggestion.
        article_entry = Article.objects.create(
            title="September", slug="september", body="During September..."
        )
        form = django_comments.get_form()(article_entry)
        data.update(form.initial)
        post_article_comment(data, article_entry, auth_user=self.user)
        self.cmts = InkComment.objects.all()

    def test_anonymous_user_redirected_when_flagging(self):
        comment = django_comments.get_model().objects.for_app_models(
            "tests.diary"
        )[0]
        flag_url = reverse("comments-flag", args=[comment.id])
        request = request_factory.get(flag_url)
        request.user = AnonymousUser()
        response = views.flag(request, comment.id)
        dest_url = "/accounts/login/?next=/comments/flag/1/"
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, dest_url)

    def test_loggedin_user_can_flag_comment(self):
        comment = django_comments.get_model().objects.for_app_models(
            "tests.diary"
        )[0]
        flag_url = reverse("comments-flag", args=[comment.id])
        request = request_factory.get(flag_url)
        request.user = self.user
        response = views.flag(request, comment.id)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(response.content.find(b"Flag comment") > -1)
        request = request_factory.post(flag_url)
        request.user = self.user
        request._dont_enforce_csrf_checks = True
        response = views.flag(request, comment.id)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("comments-flag-done") + "?c=1")
        flags = CommentFlag.objects.filter(
            comment=comment, user=self.user, flag=CommentFlag.SUGGEST_REMOVAL
        )
        self.assertTrue(flags.count() == 1)

    def test_email_is_triggered(self):
        flag_url = reverse("comments-flag", args=[1])
        self.assertTrue(self.mailer.call_count == 0)
        request = request_factory.post(flag_url)
        request.user = self.user
        request._dont_enforce_csrf_checks = True
        views.flag(request, 1)
        self.assertTrue(self.mailer.call_count == 1)

    def test_no_email_when_notify_removal_suggestion_is_None(self):
        with patch.object(
            DiaryCommentModerator, "removal_suggestion_notification", None
        ):
            flag_url = reverse("comments-flag", args=[1])
            self.assertTrue(self.mailer.call_count == 0)
            request = request_factory.post(flag_url)
            request.user = self.user
            request._dont_enforce_csrf_checks = True
            views.flag(request, 1)
            self.assertTrue(self.mailer.call_count == 0)

    @patch.multiple(
        "django_comments_ink.conf.settings",
        COMMENTS_INK_APP_MODEL_OPTIONS=app_model_options_mock,
    )
    def test_no_email_when_flagging_a_comment_sent_to_an_article(self):
        # The comment with pk=2 has been sent to an article,
        # and the Article model has no moderation registered in
        # (test/models.py), so flagging the comment for removal
        # will not trigger the notify_removal_suggestion.
        flag_url = reverse("comments-flag", args=[2])
        self.assertTrue(self.mailer.call_count == 0)
        request = request_factory.post(flag_url)
        request.user = self.user
        request._dont_enforce_csrf_checks = True
        views.flag(request, 2)
        self.assertTrue(self.mailer.call_count == 0)

    def test_no_email_when_flag_is_other_than_suggest_removal(self):
        # Here the view is django_comments.views.delete, which
        # flags a comment with MODERATION_DELETION. So no email
        # is sent as notify_removal_suggestion requires that the flag
        # is REMOVAL_SUGGESTION.
        delete_url = reverse("comments-delete", args=[1])
        self.assertTrue(self.mailer.call_count == 0)
        # The user sending MODERATION_DELETION requires permission:
        # django_comments.can_moderate.
        ct_comments = ContentType.objects.get_for_model(Comment)
        can_moderate = Permission.objects.get(
            codename="can_moderate", content_type=ct_comments
        )
        self.user.user_permissions.add(can_moderate)
        # The rest is as in previous test methods.
        request = request_factory.post(delete_url)
        request.user = self.user
        request._dont_enforce_csrf_checks = True
        delete(request, 1)
        self.assertTrue(self.mailer.call_count == 0)



from datetime import datetime
import pytest

from django.contrib.auth.models import AnonymousUser, User
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.test import RequestFactory, TestCase

import django_comments

from django_comments_ink import (
    caching,
    get_comment_reactions_enum,
    get_model,
    get_object_reactions_enum,
)
from django_comments_ink.conf import settings
from django_comments_ink.models import CommentReaction, ObjectReaction

from django_comments_ink.tests.models import Article, Diary
from django_comments_ink.tests.test_views import (
    post_article_comment,
    post_diary_comment,
)
from django_comments_ink.tests.utils import post_flag, send_reaction


request_factory = RequestFactory()
comments_model = django_comments.get_model()


class HTTPMethodsNotAllowedTests(TestCase):
    def setUp(self):
        diary_entry = Diary.objects.create(
            body="What I did in October...",
            allow_comments=True,
            publish=datetime.now(),
        )
        form = django_comments.get_form()(diary_entry)
        self.user = User.objects.create_user("bob", "bob@example.com", "pwd")
        data = {
            "name": "Bob",
            "email": "bob@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmail eine kleine...",
        }
        data.update(form.initial)
        post_diary_comment(data, diary_entry, auth_user=self.user)
        self.comment = get_model().objects.first()
        self.renum = get_comment_reactions_enum()

    def test_send_delete_method_raises_405_method_not_allowed(self):
        data = {"comment": self.comment.id, "reaction": self.renum.LIKE_IT}
        response = send_reaction("delete", data, auth_user=self.user)
        self.assertEqual(response.status_code, 405)

    def test_send_get_method_raises_405_method_not_allowed(self):
        data = {"comment": self.comment.id, "reaction": self.renum.LIKE_IT}
        response = send_reaction("get", data, auth_user=self.user)
        self.assertEqual(response.status_code, 405)

    def test_send_put_method_raises_405_method_not_allowed(self):
        data = {"comment": self.comment.id, "reaction": self.renum.LIKE_IT}
        response = send_reaction("put", data, auth_user=self.user)
        self.assertEqual(response.status_code, 405)

    def test_send_update_method_raises_405_method_not_allowed(self):
        data = {"comment": self.comment.id, "reaction": self.renum.LIKE_IT}
        response = send_reaction("put", data, auth_user=self.user)
        self.assertEqual(response.status_code, 405)


class AllowedCommentReactionTests(TestCase):
    """Scenario to test like and dislike reactions on tests.diary model."""

    def setUp(self):
        diary_entry = Diary.objects.create(
            body="What I did in October...",
            allow_comments=True,
            publish=datetime.now(),
        )
        form = django_comments.get_form()(diary_entry)
        self.user = User.objects.create_user("bob", "bob@example.com", "pwd")
        data = {
            "name": "Bob",
            "email": "bob@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmail eine kleine...",
        }
        data.update(form.initial)
        post_diary_comment(data, diary_entry, auth_user=self.user)
        self.comment = get_model().objects.first()
        self.renum = get_comment_reactions_enum()

    def test_post_reaction_as_anonymous_user_results_in_403(self):
        data = {"comment": self.comment.id, "reaction": self.renum.LIKE_IT}
        anonymous_user = AnonymousUser()
        response = send_reaction("post", data, auth_user=anonymous_user)
        self.assertEqual(response.status_code, 403)

    def test_create_LIKE_IT_reaction(self):
        data = {"comment": self.comment.id, "reaction": self.renum.LIKE_IT}
        response = send_reaction("post", data, auth_user=self.user)
        # Verify the content of the response.data attribute.
        self.assertTrue("counter" in response.data)  # Total # of reactions.
        self.assertTrue("list" in response.data)  # The list of reactions.
        item = response.data["list"][0]
        self.assertEqual(item["value"], "+")
        self.assertEqual(item["authors"][0], "bob")  # 1st item of authors.
        self.assertEqual(item["counter"], 1)  # Count of LIKE_IT reactions.
        self.assertEqual(item["label"], "+1")
        self.assertEqual(item["icon"], "#128077")
        # Verify the status of the response.
        self.assertEqual(response.status_code, 201)

    def test_sending_same_reaction_twice_cancels_it(self):
        # The first like-it reaction creates the reaction.
        self.test_create_LIKE_IT_reaction()
        data = {"comment": self.comment.id, "reaction": self.renum.LIKE_IT}
        # Check that the counter for this reaction is 1.
        creaction = CommentReaction.objects.get(**data)
        self.assertEqual(creaction.counter, 1)
        # The second like-it removes it.
        response = send_reaction("post", data, auth_user=self.user)
        self.assertEqual(response.status_code, 200)
        with self.assertRaises(CommentReaction.DoesNotExist):
            creaction = CommentReaction.objects.get(**data)

    def test_create_DISLIKE_IT_reaction(self):
        data = {"comment": self.comment.id, "reaction": self.renum.DISLIKE_IT}
        response = send_reaction("post", data, auth_user=self.user)
        self.assertEqual(response.status_code, 201)

    def test_create_both_likeit_and_dislikeit_reactions(self):
        # Like and dislike reactions can be sent from the same user for
        # the same comment. Because anything can be liked and disliked at
        # the same time. Embrace diversity!
        self.test_create_LIKE_IT_reaction()
        self.test_create_DISLIKE_IT_reaction()
        comment_reactions = CommentReaction.objects.filter(authors=self.user)
        self.assertEqual(len(comment_reactions), 2)
        crec1, crec2 = comment_reactions
        self.assertEqual(crec1.comment, self.comment)
        self.assertEqual(crec1.reaction, self.renum.LIKE_IT)
        self.assertEqual(crec2.comment, self.comment)
        self.assertEqual(crec2.reaction, self.renum.DISLIKE_IT)

    def test_create_LIKE_IT_reaction_from_different_user(self):
        # First I send the LIKE_IT reaction from above, to create the
        # CommentReaction and have the counter already to 1.
        self.test_create_LIKE_IT_reaction()
        data = {"comment": self.comment.id, "reaction": self.renum.LIKE_IT}
        comment_reaction = CommentReaction.objects.get(**data)
        self.assertEqual(comment_reaction.counter, 1)

        alice = User.objects.create_user("alice", "alice@tal.net", "pwd")
        response = send_reaction("post", data, auth_user=alice)
        self.assertEqual(response.status_code, 201)

        comment_reaction = CommentReaction.objects.get(**data)
        self.assertEqual(comment_reaction.counter, 2)

    def test_create_two_LIKE_IT_reaction_and_remove_one(self):
        self.test_create_LIKE_IT_reaction()
        data = {"comment": self.comment.id, "reaction": self.renum.LIKE_IT}
        cm_rct_1 = CommentReaction.objects.get(**data)
        self.assertEqual(cm_rct_1.counter, 1)
        alice = User.objects.create_user("alice", "alice@tal.net", "pwd")
        response = send_reaction("post", data, auth_user=alice)
        self.assertEqual(response.status_code, 201)
        cm_rct_2 = CommentReaction.objects.get(**data)
        self.assertEqual(cm_rct_2.counter, 2)
        response = send_reaction("post", data, auth_user=self.user)
        self.assertEqual(response.status_code, 200)
        cm_rct_3 = CommentReaction.objects.get(**data)
        self.assertEqual(cm_rct_3.counter, 1)
        self.assertTrue(cm_rct_1.pk == cm_rct_2.pk == cm_rct_3.pk)


class DisallowedCommentReactionTests(TestCase):
    """Scenario to test posting reactions in a disallowed scenario."""

    def setUp(self):
        article_entry = Article.objects.create(
            title="This is the title of the article",
            slug="title-article",
            body="This is the body of the article, blah blah blah...",
            allow_comments=True,
            publish=datetime.now(),
        )
        form = django_comments.get_form()(article_entry)
        self.user = User.objects.create_user("bob", "bob@example.com", "pwd")
        data = {
            "name": "Bob",
            "email": "bob@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmail eine kleine...",
        }
        data.update(form.initial)
        post_article_comment(data, article_entry, auth_user=self.user)
        self.comment = get_model().objects.first()
        self.renum = get_comment_reactions_enum()

    def test_post_reaction_as_anonymous_user_results_in_403(self):
        # This test is also in the class above: AllowedCommentReactionTests.
        # And it is exactly the same as in both cases, whether the
        # 'comment_reactions_enabled' flag is True or False, the
        # authentication is checked before hand.
        data = {"comment": self.comment.id, "reaction": self.renum.LIKE_IT}
        anonymous_user = AnonymousUser()
        response = send_reaction("post", data, auth_user=anonymous_user)
        self.assertEqual(response.status_code, 403)

    def test_create_LIKE_IT_reaction(self):
        # This test is also in the class above: AllowedCommentReactionTests.
        # As 'comment_reactions_enabled' is False for model 'tests.article',
        # posting a reaction to a comment posted to a 'tests.article' results
        # in a HTTP 403 (No permission).
        data = {"comment": self.comment.id, "reaction": self.renum.LIKE_IT}
        response = send_reaction("post", data, auth_user=self.user)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(CommentReaction.objects.filter(**data).count(), 0)

    def test_create_DISLIKE_IT_reaction(self):
        # This test is also in the class above: AllowedCommentReactionTests.
        # As 'comment_reactions_enabled' is False for model 'tests.article',
        # posting a reaction to a comment posted to a 'tests.article' results
        # in a HTTP 403 (No permission).
        data = {"comment": self.comment.id, "reaction": self.renum.DISLIKE_IT}
        response = send_reaction("post", data, auth_user=self.user)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(CommentReaction.objects.filter(**data).count(), 0)


# This class tests flagging comments posted to a diary entry.
# The model 'tests.diary' has the 'comment_flagging_enabled' = True in
# tests.settings, so flagging is allowed.
class AllowedCreateReportFlagTests(TestCase):
    def setUp(self):
        diary_entry = Diary.objects.create(
            body="What I did in October...",
            allow_comments=True,
            publish=datetime.now(),
        )
        form = django_comments.get_form()(diary_entry)
        self.user = User.objects.create_user("bob", "bob@example.com", "pwd")
        data = {
            "name": "Bob",
            "email": "bob@example.com",
            "followup": True,
            "reply_to": 0,
            "level": 1,
            "order": 1,
            "comment": "Es war einmail eine kleine...",
        }
        data.update(form.initial)
        post_diary_comment(data, diary_entry, auth_user=self.user)
        self.comment = get_model().objects.first()

    def test_flag_as_anonymous_user_results_in_403(self):
        data = {"comment": self.comment.id, "flag": "report"}
        anonymous_user = AnonymousUser()
        response = post_flag(data, auth_user=anonymous_user)
        self.assertEqual(response.status_code, 403)

    def test_flag_as_logged_in_user(self):
        data = {"comment": self.comment.id, "flag": "report"}
        response = post_flag(data, auth_user=self.user)
        self.assertEqual(response.status_code, 201)


# ---------------------------------------------------------------------
# Test ObjectReaction class.


@pytest.mark.django_db
def test_create_object_reaction_for_an_article(an_article):
    obj_reactions_enum = get_object_reactions_enum()
    ct = ContentType.objects.get_for_model(an_article)
    site = Site.objects.get(pk=1)
    object_reaction = ObjectReaction.objects.create(
        reaction=obj_reactions_enum.LIKE_IT,
        content_type=ct,
        object_pk=an_article.id,
        site=site,
    )
    assert object_reaction.counter == 0
    assert object_reaction.authors.count() == 0


@pytest.mark.django_db
def test_object_reaction_is_deleted_from_cache(an_user, an_article):
    obj_reactions_enum = get_object_reactions_enum()
    ct = ContentType.objects.get_for_model(an_article)
    site = Site.objects.get(pk=1)
    object_reaction = ObjectReaction.objects.create(
        reaction=obj_reactions_enum.LIKE_IT,
        content_type=ct,
        object_pk=an_article.id,
        site=site,
    )
    assert object_reaction.counter == 0
    assert object_reaction.authors.count() == 0

    # Create key and put it in the cache.
    dci_cache = caching.get_cache()
    key = settings.COMMENTS_INK_CACHE_KEYS["object_reactions"].format(
        ctype_pk=ct.id, object_pk=an_article.id, site_id=site.id
    )
    dci_cache.set(key, "nothing-in-particular")
    assert dci_cache.get(key) != None

    # Add user and increment counter.
    object_reaction.authors.add(an_user)
    object_reaction.counter += 1
    object_reaction.save()  # It deletes the key on save.
    assert dci_cache.get(key) == None



from django import forms
from django.apps import apps
from django.utils.translation import gettext_lazy as _
from django_comments.forms import CommentForm
from django_comments_ink.conf import settings
from django_comments_ink.models import TmpInkComment


class InkCommentForm(CommentForm):
    followup = forms.BooleanField(
        required=False, label=_("Notify me about follow-up comments")
    )
    reply_to = forms.IntegerField(
        required=True, initial=0, widget=forms.HiddenInput()
    )

    def __init__(self, *args, **kwargs):
        comment = kwargs.pop("comment", None)
        if comment:
            initial = kwargs.pop("initial", {})
            initial.update({"reply_to": comment.pk})
            kwargs["initial"] = initial
            followup_suffix = "_%d" % comment.pk
        else:
            followup_suffix = ""

        super(CommentForm, self).__init__(*args, **kwargs)

        self.fields["name"].label = _("Name")
        self.fields["name"].widget = forms.TextInput(
            attrs={"placeholder": _("name"), "class": "form-control"}
        )

        self.fields["email"].label = _("Mail")
        self.fields["email"].help_text = _("Required for comment verification")
        self.fields["email"].widget = forms.TextInput(
            attrs={"placeholder": _("mail address"), "class": "form-control"}
        )

        self.fields["url"].label = _("Link")
        self.fields["url"].required = False
        self.fields["url"].widget = forms.TextInput(
            attrs={
                "placeholder": _("url your name links to (optional)"),
                "class": "form-control",
            }
        )

        self.fields["comment"].widget = forms.Textarea(
            attrs={"placeholder": _("Your comment"), "class": "form-control"}
        )
        self.fields["comment"].max_length = settings.COMMENT_MAX_LENGTH
        self.fields["comment"].widget.attrs.pop("cols")
        self.fields["comment"].widget.attrs.pop("rows")

        self.fields["followup"].widget.attrs["id"] = (
            "id_followup%s" % followup_suffix
        )
        self.fields["followup"].widget.attrs["class"] = "custom-control-input"
        self.fields["followup"].initial = settings.COMMENTS_INK_DEFAULT_FOLLOWUP

    def get_comment_model(self):
        return TmpInkComment

    def get_comment_create_data(self, site_id=None):
        data = super(CommentForm, self).get_comment_create_data(site_id=site_id)
        ctype = data.get("content_type")
        object_pk = data.get("object_pk")
        model = apps.get_model(ctype.app_label, ctype.model)
        target = model._default_manager.get(pk=object_pk)
        data.update(
            {
                "thread_id": 0,
                "level": 0,
                "order": 1,
                "parent_id": self.cleaned_data["reply_to"],
                "followup": self.cleaned_data["followup"],
                "content_object": target,
            }
        )
        return data



from __future__ import unicode_literals

import os

PRJ_PATH = os.path.abspath(os.path.curdir)

DEBUG = True

ADMINS = (("Joe, the Administrator", "joe.bloggs@example.com"),)

MANAGERS = ADMINS

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "django_comments_ink",
        "USER": "",
        "PASSWORD": "",
        "HOST": "",
        "PORT": "",
    }
}

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# If running in a Windows environment this must be set to the same as your
# system time zone.
TIME_ZONE = "Europe/Berlin"
USE_TZ = False
USE_L10N = False

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = "en-us"

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

STATIC_URL = "/static/"

STATICFILES_FINDERS = (
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
)

# Absolute path to the directory that holds media.
# Example: "/home/media/media.lawrence.com/"
MEDIA_ROOT = ""

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash if there is a path component (optional in other cases).
# Examples: "http://media.lawrence.com", "http://example.com/media/"
MEDIA_URL = ""

# URL prefix for admin media -- CSS, JavaScript and images. Make sure to use a
# trailing slash.
# Examples: "http://foo.com/media/", "/media/".
ADMIN_MEDIA_PREFIX = "/media/"

SECRET_KEY = "v2824l&2-n+4zznbsk9c-ap5i)b3e8b+%*a=dxqlahm^%)68jn"

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
]

ROOT_URLCONF = "django_comments_ink.tests.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [
            os.path.join(os.path.dirname(__file__), "templates"),
            os.path.join(os.path.dirname(__file__), "..", "templates"),
        ],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
            ]
        },
    }
]


INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.staticfiles",
    "rest_framework",
    "django_comments_ink",
    "django_comments_ink.tests",
    "django_comments",
]

COMMENTS_APP = "django_comments_ink"

# Default primary key field type
# https://docs.djangoproject.com/en/4.0/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

DEFAULT_FROM_EMAIL = "Alice Bloggs <alice@example.com>"

COMMENTS_HIDE_REMOVED = True

COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED = True

COMMENTS_INK_CONFIRM_EMAIL = True
COMMENTS_INK_SALT = b"es-war-einmal-una-bella-princesa-in-a-beautiful-castle"
COMMENTS_INK_MAX_THREAD_LEVEL = 3
COMMENTS_INK_MAX_THREAD_LEVEL_BY_APP_MODEL = {"tests.diary": 0}

COMMENTS_INK_APP_MODEL_OPTIONS = {
    "tests.diary": {
        "comment_flagging_enabled": True,
        "comment_reactions_enabled": True,
        "object_reactions_enabled": True,
    },
    "tests.article": {
        "comment_flagging_enabled": False,
        "comment_reactions_enabled": False,
        "object_reactions_enabled": False,
    },
}

MY_DRF_AUTH_TOKEN = "08d9fd42468aebbb8087b604b526ff0821ce4525"

REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "rest_framework.authentication.SessionAuthentication",
        "django_comments_ink.tests.apiauth.APIRequestAuthentication",
    ]
}

TEST_RUNNER = "django.test.runner.DiscoverRunner"



from django.contrib.auth.models import AnonymousUser
from rest_framework import HTTP_HEADER_ENCODING, authentication, exceptions


class APIRequestAuthentication(authentication.BaseAuthentication):
    def authenticate(self, request):
        auth = request.META.get("HTTP_AUTHORIZATION", b"")
        if isinstance(auth, str):
            auth = auth.encode(HTTP_HEADER_ENCODING)

        pieces = auth.split()
        if not pieces or pieces[0].lower() != b"token":
            return None

        if len(pieces) == 1:
            msg = _("Invalid token header. No credentials provided.")
            raise exceptions.AuthenticationFailed(msg)
        elif len(pieces) > 2:
            msg = _(
                "Invalid token header."
                "Token string should not contain spaces."
            )
            raise exceptions.AuthenticationFailed(msg)

        try:
            auth = pieces[1].decode()
        except UnicodeError:
            msg = _(
                "Invalid token header. "
                "Token string should not contain invalid characters."
            )

        return (AnonymousUser(), auth)



from __future__ import unicode_literals

from django.apps import apps
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.decorators import login_required
from django.contrib.auth.views import redirect_to_login
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.views import shortcut
from django.contrib.sites.shortcuts import get_current_site
from django.core import signing
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.core.paginator import Paginator
from django.db.models import F
from django.db.utils import NotSupportedError
from django.http import (
    Http404,
    HttpResponseForbidden,
    HttpResponseRedirect,
    JsonResponse,
)
from django.shortcuts import get_object_or_404, render, resolve_url
from django.template import loader
from django.urls import reverse
from django.utils.html import escape
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_protect
from django.views.decorators.http import require_POST
from django.views.defaults import bad_request

from django_comments.signals import comment_was_posted, comment_will_be_posted
from django_comments.views.comments import CommentPostBadRequest
from django_comments.views.moderation import perform_flag
from django_comments.views.utils import next_redirect

from django_comments_ink import (
    get_comment_reactions_enum,
    get_form,
    get_object_reactions_enum,
)
from django_comments_ink import get_model as get_comment_model
from django_comments_ink import signals, signed, utils
from django_comments_ink.conf import settings
from django_comments_ink.models import (
    CommentReaction,
    MaxThreadLevelExceededException,
    ObjectReaction,
    TmpInkComment,
)


InkComment = get_comment_model()


if len(settings.COMMENTS_INK_THEME_DIR) > 0:
    theme_dir = f"themes/{settings.COMMENTS_INK_THEME_DIR}"
    theme_dir_exists = utils.does_theme_dir_exist(f"comments/{theme_dir}")
else:
    theme_dir = ""
    theme_dir_exists = False


# List of possible paths to the bad_form.html template file.
if theme_dir_exists:
    _bad_form_tmpl = [
        f"comments/{theme_dir}/bad_form.html",
        "comments/bad_form.html",
    ]
else:
    _bad_form_tmpl = "comments/bad_form.html"


# List of possible paths to the preview.html template file.
_preview_tmpl = []
if theme_dir_exists:
    _preview_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/preview.html",
            "comments/{theme_dir}/{app_label}/preview.html",
            "comments/{theme_dir}/preview.html",
        ]
    )
_preview_tmpl.extend(
    [
        "comments/{app_label}/{model}/preview.html",
        "comments/{app_label}/preview.html",
        "comments/preview.html",
    ]
)


# List of possible paths to the 'reply_form_js.html' or 'form.js.html'.
_preview_js_tmpl = []
if theme_dir_exists:
    _preview_js_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/{template_name}",
            "comments/{theme_dir}/{app_label}/{template_name}",
            "comments/{theme_dir}/{template_name}",
        ]
    )
_preview_js_tmpl.extend(
    [
        "comments/{app_label}/{model}/{template_name}",
        "comments/{app_label}/{template_name}",
        "comments/{template_name}",
    ]
)


# List of possible paths to the bad_form.html template file.
if theme_dir_exists:
    _posted_tmpl = [f"comments/{theme_dir}/posted.html", "comments/posted.html"]
else:
    _posted_tmpl = "comments/posted.html"


# List of possible paths to the moderated.html template file.
_moderated_tmpl = []
if theme_dir_exists:
    _moderated_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/moderated.html",
            "comments/{theme_dir}/{app_label}/moderated.html",
            "comments/{theme_dir}/moderated.html",
        ]
    )
_moderated_tmpl.extend(
    [
        "comments/{app_label}/{model}/moderated.html",
        "comments/{app_label}/moderated.html",
        "comments/moderated.html",
    ]
)


# List of possible paths to the posted_js.html template file.
_posted_js_tmpl = []
if theme_dir_exists:
    _posted_js_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/posted_js.html",
            "comments/{theme_dir}/{app_label}/posted_js.html",
            "comments/{theme_dir}/posted_js.html",
        ]
    )
_posted_js_tmpl.extend(
    [
        "comments/{app_label}/{model}/posted_js.html",
        "comments/{app_label}/posted_js.html",
        "comments/posted_js.html",
    ]
)


# List of possible paths to the published_js.html template file.
_published_js_tmpl = []
if theme_dir_exists:
    _published_js_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/published_js.html",
            "comments/{theme_dir}/{app_label}/published_js.html",
            "comments/{theme_dir}/published_js.html",
        ]
    )
_published_js_tmpl.extend(
    [
        "comments/{app_label}/{model}/published_js.html",
        "comments/{app_label}/published_js.html",
        "comments/published_js.html",
    ]
)


# List of possible paths to the moderated_js.html template file.
_moderated_js_tmpl = []
if theme_dir_exists:
    _moderated_js_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/moderated_js.html",
            "comments/{theme_dir}/{app_label}/moderated_js.html",
            "comments/{theme_dir}/moderated_js.html",
        ]
    )
_moderated_js_tmpl.extend(
    [
        "comments/{app_label}/{model}/moderated_js.html",
        "comments/{app_label}/moderated_js.html",
        "comments/moderated_js.html",
    ]
)


# List of possible paths to the discarded.html template file.
if theme_dir_exists:
    _discarded_tmpl = [
        f"comments/{theme_dir}/discarded.html",
        "comments/discarded.html",
    ]
else:
    _discarded_tmpl = "comments/discarded.html"


# List of possible paths to the reply.html template file.
_reply_tmpl = []
if theme_dir_exists:
    _reply_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/reply.html",
            "comments/{theme_dir}/{app_label}/reply.html",
            "comments/{theme_dir}/reply.html",
        ]
    )
_reply_tmpl.extend(
    [
        "comments/{app_label}/{model}/reply.html",
        "comments/{app_label}/reply.html",
        "comments/reply.html",
    ]
)


# List of possible paths to the muted.html template file.
_muted_tmpl = []
if theme_dir_exists:
    _muted_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/muted.html",
            "comments/{theme_dir}/{app_label}/muted.html",
            "comments/{theme_dir}/muted.html",
        ]
    )
_muted_tmpl.extend(
    [
        "comments/{app_label}/{model}/muted.html",
        "comments/{app_label}/muted.html",
        "comments/muted.html",
    ]
)


# List of possible paths to the react.html template file.
if theme_dir_exists:
    _react_tmpl = [
        f"comments/{theme_dir}/react.html",
        "comments/react.html",
    ]
else:
    _react_tmpl = "comments/react.html"


# List of possible paths to the reacted_js.html template file.
_reacted_js_tmpl = []
if theme_dir_exists:
    _reacted_js_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/comment_reactions.html",
            "comments/{theme_dir}/{app_label}/comment_reactions.html",
            "comments/{theme_dir}/comment_reactions.html",
        ]
    )
_reacted_js_tmpl.extend(
    [
        "comments/{app_label}/{model}/comment_reactions.html",
        "comments/{app_label}/comment_reactions.html",
        "comments/comment_reactions.html",
    ]
)


# List of possible paths to the reacted_to_comment.html template file.
if theme_dir_exists:
    _reacted_tmpl = [
        f"comments/{theme_dir}/reacted.html",
        "comments/reacted.html",
    ]
else:
    _reacted_tmpl = "comments/reacted.html"


# List of possible paths to the list_reacted.html template file.
_list_reacted_tmpl = []
if theme_dir_exists:
    _list_reacted_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/list_reacted.html",
            "comments/{theme_dir}/{app_label}/list_reacted.html",
            "comments/{theme_dir}/list_reacted.html",
        ]
    )
_list_reacted_tmpl.extend(
    [
        "comments/{app_label}/{model}/list_reacted.html",
        "comments/{app_label}/list_reacted.html",
        "comments/list_reacted.html",
    ]
)


# List of possible paths to the list_reacted_to_object.html template file.
_list_reacted_to_object_tmpl = []
if theme_dir_exists:
    _list_reacted_to_object_tmpl.extend(
        [
            "comments/{theme_dir}/{app_label}/{model}/"
            "list_reacted_to_object.html",
            "comments/{theme_dir}/{app_label}/list_reacted_to_object.html",
            "comments/{theme_dir}/list_reacted_to_object.html",
        ]
    )
_list_reacted_to_object_tmpl.extend(
    [
        "comments/{app_label}/{model}/list_reacted_to_object.html",
        "comments/{app_label}/list_reacted_to_object.html",
        "comments/list_reacted_to_object.html",
    ]
)


# ---------------------------------------------------------------------
@csrf_protect
@require_POST
def post(request, next=None, using=None):
    """
    Post a comment.

    HTTP POST is required. If ``POST['submit'] == "preview"`` or if there are
    errors a preview template, ``comments/<theme_dir>/preview.html``, will be
    rendered.

    This function is copied from the original in django-comments. It extends
    it to check whether the comment form has a field with the name indicated in
    the setting COMMENTS_INK_PAGE_QUERY_STRING_PARAM. If so it is added to the
    context passed to the template as `page_number`, which corresponds to the
    comment's page number in which the comment has been displayed.
    """
    if request.META.get("HTTP_X_REQUESTED_WITH") == "XMLHttpRequest":
        return post_js(request, next, using)

    data = request.POST.copy()
    if request.user.is_authenticated:
        if not data.get("name", ""):
            data["name"] = (
                request.user.get_full_name() or request.user.get_username()
            )
        if not data.get("email", ""):
            data["email"] = request.user.email

    # Look up the object we're trying to comment about.
    ctype = data.get("content_type")
    object_pk = data.get("object_pk")
    if ctype is None or object_pk is None:
        return CommentPostBadRequest("Missing content_type or object_pk field.")
    try:
        model = apps.get_model(*ctype.split(".", 1))
        target = model._default_manager.using(using).get(pk=object_pk)
    except (LookupError, TypeError):
        return CommentPostBadRequest(
            "Invalid content_type value: %r" % escape(ctype)
        )
    except AttributeError:
        return CommentPostBadRequest(
            "The given content-type %r does not resolve to a valid model."
            % escape(ctype)
        )
    except ObjectDoesNotExist:
        return CommentPostBadRequest(
            "No object matching content-type %r and object PK %r exists."
            % (escape(ctype), escape(object_pk))
        )
    except (ValueError, ValidationError) as e:
        return CommentPostBadRequest(
            "Attempting to get content-type %r and object PK %r raised %s"
            % (escape(ctype), escape(object_pk), e.__class__.__name__)
        )

    # Do we want to preview the comment?
    preview = "preview" in data

    # Construct the comment form
    form = get_form()(target, data=data)

    # Check security information
    if form.security_errors():
        return CommentPostBadRequest(
            "The comment form failed security verification: %s"
            % escape(str(form.security_errors()))
        )

    cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
    cpage = request.POST.get(cpage_qs_param, None)

    # If there are errors or if we requested a preview show the comment.
    if form.errors or preview:
        template_list = [
            pth.format(
                theme_dir=theme_dir,
                app_label=model._meta.app_label,
                model=model._meta.model_name,
            )
            for pth in _preview_tmpl
        ]
        return render(
            request,
            template_list,
            {
                "comment": form.data.get("comment", ""),
                "form": form,
                "is_reply": bool(form.data.get("reply_to")),
                "next": data.get("next", next),
                "page_number": cpage,
                "dcx_theme_dir": theme_dir,
            },
        )

    # Otherwise create the comment
    comment = form.get_comment_object(site_id=get_current_site(request).id)
    comment.ip_address = request.META.get("REMOTE_ADDR", None) or None
    comment.page_number = cpage
    if request.user.is_authenticated:
        comment.user = request.user

    # Signal that the comment is about to be saved
    responses = comment_will_be_posted.send(
        sender=comment.__class__, comment=comment, request=request
    )

    for (receiver, response) in responses:
        if response is False:
            return CommentPostBadRequest(
                "comment_will_be_posted receiver %r killed the comment"
                % receiver.__name__
            )

    # Save the comment and signal that it was saved
    comment.save()
    comment_was_posted.send(
        sender=comment.__class__, comment=comment, request=request
    )
    kwargs = {
        "c": comment._get_pk_val(),
    }
    if cpage is not None:
        kwargs[settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM] = cpage
    return next_redirect(
        request, fallback=next or "comments-comment-done", **kwargs
    )


def json_res(request, template_search_list, context, status=200):
    html = loader.render_to_string(template_search_list, context, request)
    json_context = {"html": html, "reply_to": request.POST.get("reply_to", "0")}
    if "field_focus" in context:
        json_context.update({"field_focus": context["field_focus"]})
    return JsonResponse(json_context, status=status)


def post_js(request, next=None, using=None):
    """
    Handles a comment post when the request is an XMLHttpRequest.
    """
    data = request.POST.copy()
    if request.user.is_authenticated:
        if not data.get("name", ""):
            data["name"] = (
                request.user.get_full_name() or request.user.get_username()
            )
        if not data.get("email", ""):
            data["email"] = request.user.email

    # Look up the object we're trying to comment about.
    ctype = data.get("content_type")
    object_pk = data.get("object_pk")
    if ctype is None or object_pk is None:
        context = {"bad_error": "Missing content_type or object_pk field."}
        return json_res(request, _bad_form_tmpl, context, status=400)

    try:
        model = apps.get_model(*ctype.split(".", 1))
        target = model._default_manager.using(using).get(pk=object_pk)
    except (LookupError, TypeError):
        error_msg = "Invalid content_type value: %r" % escape(ctype)
        context = {"bad_error": error_msg}
        return json_res(request, _bad_form_tmpl, context, status=400)
    except AttributeError:
        error_msg = (
            "The given content-type %r does not resolve to a valid model."
            % escape(ctype)
        )
        context = {"bad_error": error_msg}
        return json_res(request, _bad_form_tmpl, context, status=400)
    except ObjectDoesNotExist:
        error_msg = (
            "No object matching content-type %r and object PK %r exists."
            % (
                escape(ctype),
                escape(object_pk),
            )
        )
        context = {"bad_error": error_msg}
        return json_res(request, _bad_form_tmpl, context, status=400)
    except (ValueError, ValidationError) as e:
        error_msg = (
            "Attempting to get content-type %r and object PK %r raised %s"
            % (
                escape(ctype),
                escape(object_pk),
                e.__class__.__name__,
            )
        )
        context = {"bad_error": error_msg}
        return json_res(request, _bad_form_tmpl, context, status=400)

    # Do we want to preview the comment?
    preview = "preview" in data

    # Construct the comment form
    form = get_form()(target, data=data)

    # Check security information
    if form.security_errors():
        error_msg = "The comment form failed security verification."
        context = {"bad_error": error_msg}
        return json_res(request, _bad_form_tmpl, context, status=400)

    cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
    cpage = request.POST.get(cpage_qs_param, None)

    # If there are errors or if we requested a preview show the comment.
    if form.errors or preview:
        # Use different template depending on whether it's a reply or not.
        if form.data.get("reply_to", "0") != "0":
            template_name = "reply_form_js.html"
        else:
            template_name = "form_js.html"

        template_list = [
            pth.format(
                theme_dir=theme_dir,
                app_label=model._meta.app_label,
                model=model._meta.model_name,
                template_name=template_name,
            )
            for pth in _preview_js_tmpl
        ]
        if form.errors:
            field_focus = [key for key in form.errors.keys()][0]
        else:
            field_focus = None

        context = {
            "display_preview": not form.errors,
            "comment": form.data.get("comment", ""),
            "form": form,
            "field_focus": field_focus,
            "is_reply": form.data.get("reply_to", "0") != "0",
            "next": data.get("next", next),
            "page_number": cpage,
        }
        return json_res(request, template_list, context, status=200)

    # Otherwise create the comment
    comment = form.get_comment_object(site_id=get_current_site(request).id)
    comment.ip_address = request.META.get("REMOTE_ADDR", None) or None
    comment.page_number = cpage
    if request.user.is_authenticated:
        comment.user = request.user

    # Signal that the comment is about to be saved
    responses = comment_will_be_posted.send(
        sender=comment.__class__, comment=comment, request=request
    )

    for (receiver, response) in responses:
        if response is False:
            if settings.DEBUG:
                error_msg = (
                    "comment_will_be_posted receiver %r killed the comment"
                    % receiver.__name__
                )
            else:  # pragma: no cover
                error_msg = "Your comment has been rejected."

            context = {"bad_error": error_msg}
            return json_res(request, _bad_form_tmpl, context, status=400)

    # Save the comment and signal that it was saved
    comment.save()
    comment_was_posted.send(
        sender=comment.__class__, comment=comment, request=request
    )
    return sent_js(request, comment, using=using)


def send_email_confirmation_request(
    comment,
    key,
    site,
    text_template="comments/email_confirmation_request.txt",
    html_template="comments/email_confirmation_request.html",
):
    """Send email requesting comment confirmation"""
    subject = _("comment confirmation request")
    confirmation_url = reverse(
        "comments-ink-confirm", args=[key.decode("utf-8")]
    )
    message_context = {
        "comment": comment,
        "confirmation_url": confirmation_url,
        "contact": settings.COMMENTS_INK_CONTACT_EMAIL,
        "site": site,
    }
    # prepare text message
    text_message_template = loader.get_template(text_template)
    text_message = text_message_template.render(message_context)
    if settings.COMMENTS_INK_SEND_HTML_EMAIL:
        # prepare html message
        html_message_template = loader.get_template(html_template)
        html_message = html_message_template.render(message_context)
    else:
        html_message = None

    utils.send_mail(
        subject,
        text_message,
        settings.COMMENTS_INK_FROM_EMAIL,
        [
            comment.user_email,
        ],
        html=html_message,
    )


def _get_comment_if_exists(comment):
    """
    True if exists a InkComment with same user_name, user_email and submit_date.
    """
    return InkComment.objects.filter(
        user_name=comment.user_name,
        user_email=comment.user_email,
        followup=comment.followup,
        submit_date=comment.submit_date,
    ).first()


def _create_comment(tmp_comment):
    """
    Creates a InkComment from a TmpInkComment.
    """
    tmp_comment.pop("page_number", None)
    comment = InkComment(**tmp_comment)
    comment.save()
    return comment


def on_comment_will_be_posted(sender, comment, request, **kwargs):
    """
    Check whether there are conditions to reject the post.

    Returns False if there are conditions to reject the comment.
    """
    if settings.COMMENTS_APP != "django_comments_ink":  # pragma: no cover
        # Do not kill the post if handled by other commenting app.
        return True

    if comment.user:
        user_is_authenticated = comment.user.is_authenticated
    else:
        user_is_authenticated = False

    ct = comment.get("content_type")
    ct_str = "%s.%s" % (ct.app_label, ct.model)
    options = utils.get_app_model_options(content_type=ct_str)
    if not user_is_authenticated and options["who_can_post"] == "users":
        # Reject comment.
        return False
    else:
        return True


comment_will_be_posted.connect(on_comment_will_be_posted, sender=TmpInkComment)


def on_comment_was_posted(sender, comment, request, **kwargs):
    """
    Post the comment if a user is authenticated or send a confirmation email.

    On signal django_comments.signals.comment_was_posted check if the
    user is authenticated or if settings.COMMENTS_INK_CONFIRM_EMAIL is False.
    In both cases will post the comment. Otherwise will send a confirmation
    email to the person who posted the comment.
    """
    if settings.COMMENTS_APP != "django_comments_ink":  # pragma: no cover
        return False
    if comment.user:
        user_is_authenticated = comment.user.is_authenticated
    else:
        user_is_authenticated = False

    if not settings.COMMENTS_INK_CONFIRM_EMAIL or user_is_authenticated:
        if _get_comment_if_exists(comment) is None:
            new_comment = _create_comment(comment)
            comment.ink_comment = new_comment
            signals.confirmation_received.send(
                sender=TmpInkComment, comment=comment, request=request
            )
            if comment.is_public:
                notify_comment_followers(new_comment)
    else:
        key = signed.dumps(
            comment, compress=True, extra_key=settings.COMMENTS_INK_SALT
        )
        site = get_current_site(request)
        send_email_confirmation_request(comment, key, site)


comment_was_posted.connect(on_comment_was_posted, sender=TmpInkComment)


def sent(request, using=None):
    comment_pk = request.GET.get("c", None)
    if not comment_pk:
        return CommentPostBadRequest("Comment doesn't exist")
    try:
        comment_pk = int(comment_pk)
        comment = InkComment.objects.get(pk=comment_pk)
    except (TypeError, ValueError, InkComment.DoesNotExist):
        try:
            value = signing.loads(comment_pk)
            ctype, object_pk = value.split(":")
            model = apps.get_model(*ctype.split(".", 1))
            target = model._default_manager.using(using).get(pk=object_pk)
        except Exception:
            return CommentPostBadRequest("Comment doesn't exist")

        return render(request, _posted_tmpl, {"target": target})
    else:
        if comment.is_public:
            return utils.redirect_to(comment, request=request)
        else:
            template_list = [
                pth.format(
                    theme_dir=theme_dir,
                    app_label=comment.content_type.app_label,
                    model=comment.content_type.model,
                )
                for pth in _moderated_tmpl
            ]
            return render(request, template_list, {"comment": comment})


def sent_js(request, comment, using=None):
    try:
        comment_pk = comment._get_pk_val()
        comment = InkComment.objects.get(pk=comment_pk)
    except (TypeError, ValueError, InkComment.DoesNotExist):
        try:
            value = signing.loads(comment_pk)
            ctype, object_pk = value.split(":")
            model = apps.get_model(*ctype.split(".", 1))
            target = model._default_manager.using(using).get(pk=object_pk)
        except Exception:
            context = {"bad_error": "Comment does not exist."}
            return json_res(request, _bad_form_tmpl, context, status=400)
        app_label, model_name = ctype.split(".", 1)
        template_list = [
            pth.format(
                theme_dir=theme_dir, app_label=app_label, model=model_name
            )
            for pth in _posted_js_tmpl
        ]
        return json_res(request, template_list, {"target": target}, status=202)
    else:
        if comment.is_public:
            # Return a render instead of a redirect_to. But use status=201.
            # In comment_form.js check whether the status is 201 to read
            # the content as the redirect_url.
            template_list = [
                pth.format(
                    theme_dir=theme_dir,
                    app_label=comment.content_type.app_label,
                    model=comment.content_type.model,
                )
                for pth in _published_js_tmpl
            ]
            page_number = utils.get_comment_page_number(
                request, comment.content_type.id, comment.object_pk, comment.id
            )
            comment_url = utils.get_comment_url(comment, None, page_number)
            return json_res(
                request,
                template_list,
                {"comment": comment, "comment_url": comment_url},
                status=201,
            )
        else:
            template_list = [
                pth.format(
                    theme_dir=theme_dir,
                    app_label=comment.content_type.app_label,
                    model=comment.content_type.model,
                )
                for pth in _moderated_js_tmpl
            ]
            return json_res(
                request, template_list, {"comment": comment}, status=201
            )


def confirm(request, key, template_discarded=_discarded_tmpl):
    try:
        tmp_comment = signed.loads(
            str(key), extra_key=settings.COMMENTS_INK_SALT
        )
    except (ValueError, signed.BadSignature) as exc:
        return bad_request(request, exc)

    # The comment does exist if the URL was already confirmed,
    # in such a case, as per suggested in ticket #80, we return
    # the comment's URL, as if the comment is just confirmed.
    comment = _get_comment_if_exists(tmp_comment)
    page_number = tmp_comment.pop("page_number", None)

    if comment is not None:
        return utils.redirect_to(comment, page_number=page_number)

    # Send signal that the comment confirmation has been received.
    responses = signals.confirmation_received.send(
        sender=TmpInkComment, comment=tmp_comment, request=request
    )
    # Check whether a signal receiver decides to discard the comment.
    for (receiver, response) in responses:
        if response is False:
            return render(request, template_discarded, {"comment": tmp_comment})

    comment = _create_comment(tmp_comment)
    if comment.is_public is False:
        template_list = [
            pth.format(
                theme_dir=theme_dir,
                app_label=comment.content_type.app_label,
                model=comment.content_type.model,
            )
            for pth in _moderated_tmpl
        ]
        return render(request, template_list, {"comment": comment})
    else:
        notify_comment_followers(comment)
        return utils.redirect_to(comment, page_number=page_number)


def notify_comment_followers(comment):
    followers = {}
    kwargs = {
        "content_type": comment.content_type,
        "object_pk": comment.object_pk,
        "is_public": True,
        "followup": True,
    }
    previous_comments = InkComment.objects.filter(**kwargs).exclude(
        user_email=comment.user_email
    )

    def feed_followers(gen):
        for instance in gen:
            followers[instance.user_email] = (
                instance.user_name,
                signed.dumps(
                    instance,
                    compress=True,
                    extra_key=settings.COMMENTS_INK_SALT,
                ),
            )

    try:
        gen = previous_comments.distinct("user_email").order_by("user_email")
        feed_followers(gen)
    except NotSupportedError:
        feed_followers(previous_comments)

    for instance in previous_comments:
        followers[instance.user_email] = (
            instance.user_name,
            signed.dumps(
                instance, compress=True, extra_key=settings.COMMENTS_INK_SALT
            ),
        )

    subject = _("new comment posted")
    text_message_template = loader.get_template(
        "comments/email_followup_comment.txt"
    )
    if settings.COMMENTS_INK_SEND_HTML_EMAIL:
        html_message_template = loader.get_template(
            "comments/email_followup_comment.html"
        )

    for email, (name, key) in followers.items():
        mute_url = reverse("comments-ink-mute", args=[key.decode("utf-8")])
        message_context = {
            "user_name": name,
            "comment": comment,
            # 'content_object': target,
            "mute_url": mute_url,
            "site": comment.site,
        }
        text_message = text_message_template.render(message_context)
        if settings.COMMENTS_INK_SEND_HTML_EMAIL:
            html_message = html_message_template.render(message_context)
        else:
            html_message = None
        utils.send_mail(
            subject,
            text_message,
            settings.COMMENTS_INK_FROM_EMAIL,
            [
                email,
            ],
            html=html_message,
        )


def reply(request, cid):
    try:
        comment = InkComment.objects.get(pk=cid)
        if not comment.allow_thread():
            raise MaxThreadLevelExceededException(comment)
    except MaxThreadLevelExceededException as exc:
        return HttpResponseForbidden(exc)
    except InkComment.DoesNotExist as exc:
        raise Http404(exc)

    ct_str = "%s.%s" % (
        comment.content_type.app_label,
        comment.content_type.model,
    )
    options = utils.get_app_model_options(content_type=ct_str)

    if not request.user.is_authenticated and options["who_can_post"] == "users":
        path = request.build_absolute_uri()
        resolved_login_url = resolve_url(settings.LOGIN_URL)
        return redirect_to_login(path, resolved_login_url, REDIRECT_FIELD_NAME)

    form = get_form()(comment.content_object, comment=comment)
    next = request.GET.get("next", reverse("comments-ink-sent"))

    cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
    cpage = request.POST.get(cpage_qs_param, None)
    cfold_qs_param = settings.COMMENTS_INK_FOLD_QUERY_STRING_PARAM
    cfold = request.GET.get(cfold_qs_param, "")

    template_list = [
        pth.format(
            theme_dir=theme_dir,
            app_label=comment.content_type.app_label,
            model=comment.content_type.model,
        )
        for pth in _reply_tmpl
    ]
    return render(
        request,
        template_list,
        {
            "comment": comment,
            "form": form,
            "cid": cid,
            "next": next,
            "page_number": cpage,
            "folded_comments": cfold,
        },
    )


def mute(request, key):
    try:
        tmp_comment = signed.loads(
            str(key), extra_key=settings.COMMENTS_INK_SALT
        )
    except (ValueError, signed.BadSignature) as exc:
        return bad_request(request, exc)

    # Can't mute a comment that doesn't have the followup attribute
    # set to True, or a comment that doesn't exist.
    if not tmp_comment.followup or _get_comment_if_exists(tmp_comment) is None:
        raise Http404(_("Comment already muted or comment does not exist."))

    # Send signal that the comment thread has been muted
    signals.comment_thread_muted.send(
        sender=InkComment, comment=tmp_comment, request=request
    )

    InkComment.norel_objects.filter(
        content_type=tmp_comment.content_type,
        object_pk=tmp_comment.object_pk,
        user_email=tmp_comment.user_email,
        is_public=True,
        followup=True,
    ).update(followup=False)

    model = apps.get_model(
        tmp_comment.content_type.app_label, tmp_comment.content_type.model
    )
    target = model._default_manager.get(pk=tmp_comment.object_pk)
    template_list = [
        pth.format(
            theme_dir=theme_dir,
            app_label=tmp_comment.content_type.app_label,
            model=tmp_comment.content_type.model,
        )
        for pth in _muted_tmpl
    ]
    return render(request, template_list, {"content_object": target})


@csrf_protect
@login_required
def flag(request, comment_id, next=None):
    """
    Flags a comment. Confirmation on GET, action on POST.

    Templates: :template:`comments/flag.html`,
    Context:
        comment_id
            The id of the comment the user is flagging.
    """
    comment = get_object_or_404(
        get_comment_model(),
        pk=comment_id,
        site__pk=utils.get_current_site_id(request),
    )
    utils.check_option("comment_flagging_enabled", comment=comment)

    # Flag on POST.
    if request.method == "POST":
        perform_flag(request, comment)
        return next_redirect(
            request, fallback=next or "comments-flag-done", c=comment.pk
        )

    # Render a form on GET
    else:
        return render(
            request,
            "comments/flag.html",
            {
                "comment": comment,
                "next": next,
            },
        )


@csrf_protect
@login_required
def react(request, comment_id, next=None):
    """
    A registered user reacts to a comment. Confirmation on GET, action on POST.

    Templates: :template:`comments/<theme_dir>/react.html`,
    Context:
        comment
            the `comments.comment` object the user reacted to.
    """
    comment = get_object_or_404(
        get_comment_model(),
        pk=comment_id,
        site__pk=utils.get_current_site_id(request),
    )
    utils.check_option("comment_reactions_enabled", comment=comment)
    cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
    cfold_qs_param = settings.COMMENTS_INK_FOLD_QUERY_STRING_PARAM

    if request.method == "POST":
        cpage = request.POST.get(cpage_qs_param, 1)
        cfold = request.POST.get(cfold_qs_param, "")

        created = perform_react(request, comment)

        # When the reaction has been sent via JavaScript.
        if request.META.get("HTTP_X_REQUESTED_WITH") == "XMLHttpRequest":
            obj_meta = comment.content_object._meta
            template_list = [
                pth.format(
                    theme_dir=theme_dir,
                    app_label=obj_meta.app_label,
                    model=obj_meta.model_name,
                )
                for pth in _reacted_js_tmpl
            ]
            max_users_in_tooltip = settings.COMMENTS_INK_MAX_USERS_IN_TOOLTIP
            context = {
                "comment": comment,
                "max_users_in_tooltip": max_users_in_tooltip,
            }
            status = 201 if created else 200
            return json_res(request, template_list, context, status=status)

        page = int(cpage) if cpage != "last" else "last"
        fold = (len(cfold) and {int(cid) for cid in cfold.split(",")}) or {}

        kwargs = {
            "c": comment.pk,
            cpage_qs_param: page,
            cfold_qs_param: ",".join([str(cid) for cid in fold]),
        }

        return next_redirect(
            request,
            fallback=next or "comments-ink-react-done",
            **kwargs,
        )

    else:
        cpage = request.GET.get(cpage_qs_param, 1)
        cfold = request.GET.get(cfold_qs_param, "")

        page = int(cpage) if cpage != "last" else "last"
        fold = (len(cfold) and {int(cid) for cid in cfold.split(",")}) or {}

        user_reactions = []
        cr_qs = CommentReaction.objects.filter(
            comment=comment, authors=request.user
        )
        for cmt_reaction in cr_qs:
            user_reactions.append(
                get_comment_reactions_enum()(cmt_reaction.reaction)
            )

        return render(
            request,
            _react_tmpl,
            {
                "comment": comment,
                "user_reactions": user_reactions,
                "next": next,
                "page_number": page,
                "folded_comments": ",".join([str(cid) for cid in fold]),
                "comments_page_qs_param": cpage_qs_param,
                "comments_fold_qs_param": cfold_qs_param,
            },
        )


def perform_react(request, comment):
    """Save the user reaction and send the signal comment_got_a_reaction."""
    created = False
    cr_qs = CommentReaction.objects.filter(
        reaction=request.POST["reaction"], comment=comment
    )
    if cr_qs.filter(authors=request.user).count() == 1:
        if cr_qs[0].counter == 1:
            cr_qs[0].delete_from_cache()
            cr_qs.delete()
        else:
            cr_qs.update(counter=F("counter") - 1)
            cr_qs[0].delete_from_cache()
            cr_qs[0].authors.remove(request.user)
    else:
        cmt_reaction, created = CommentReaction.objects.get_or_create(
            reaction=request.POST["reaction"], comment=comment
        )
        cmt_reaction.authors.add(request.user)
        cmt_reaction.counter += 1
        cmt_reaction.save()
    signals.comment_got_a_reaction.send(
        sender=comment.__class__,
        comment=comment,
        reaction=request.POST["reaction"],
        created=created,
        request=request,
    )
    return created


def react_done(request):
    """Displays a "User reacted to this comment" success page."""
    comment_pk = request.GET.get("c", None)
    cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
    cpage = request.GET.get(cpage_qs_param, 1)
    cfold_qs_param = settings.COMMENTS_INK_FOLD_QUERY_STRING_PARAM
    cfold = request.GET.get(cfold_qs_param, "")
    if comment_pk:
        comment = get_object_or_404(
            get_comment_model(),
            pk=comment_pk,
            site__pk=utils.get_current_site_id(request),
        )
    else:
        raise Http404(_("Comment does not exist"))

    page = int(cpage) if cpage != "last" else "last"
    fold = (len(cfold) and {int(cid) for cid in cfold.split(",")}) or {}

    return render(
        request,
        _reacted_tmpl,
        {
            "comment": comment,
            "cpage": page,
            "cfold": ",".join([str(cid) for cid in fold]),
        },
    )


@login_required
@csrf_protect
@require_POST
def react_to_object(request, content_type_id, object_pk, next=None):
    """A user reacts to an object identified with content_type and pk."""
    try:
        ctype = ContentType.objects.get(pk=content_type_id)
        ctype.get_object_for_this_type(pk=object_pk)
    except Exception:
        raise Http404(
            "Object referenced by pair (ctype_id, obj_id): (%d, %d) "
            "does not exist" % (content_type_id, object_pk)
        )

    ctype_str = ctype.app_label + "." + ctype.model
    utils.check_option("object_reactions_enabled", content_type=ctype_str)
    cpage_qs_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
    cfold_qs_param = settings.COMMENTS_INK_FOLD_QUERY_STRING_PARAM

    cpage = request.POST.get(cpage_qs_param, "")
    cfold = request.POST.get(cfold_qs_param, "")

    perform_react_to_object(request, content_type_id, object_pk)

    qs_params = []
    page = int(cpage) if not cpage in ["", "last"] else cpage
    if page:
        qs_params.append(f"{cpage_qs_param}={page}")
    fold = (len(cfold) and {int(cid) for cid in cfold.split(",")}) or {}
    if len(fold):
        qs_params.append(
            f'{cfold_qs_param}={",".join([str(cid) for cid in fold])}'
        )

    next = request.POST.get("next")

    if "#" in next:
        bits = next.rsplit("#", 1)
        next = bits[0]
        anchor = "#" + bits[1]
    else:
        anchor = ""

    if len(qs_params):
        joiner = ("?" in next) and "&" or "?"
        next += joiner + "&".join(qs_params)

    if len(anchor):
        next += anchor

    return HttpResponseRedirect(next)


def perform_react_to_object(request, content_type_id, object_pk):
    """Save the user reaction and send the signal object_got_a_reaction."""
    created = False
    ctype = ContentType.objects.get(pk=content_type_id)
    object = ctype.get_object_for_this_type(pk=object_pk)
    site = get_current_site(request)

    or_qs = ObjectReaction.objects.filter(
        reaction=request.POST["reaction"],
        content_type=ctype,
        object_pk=object_pk,
        site=site,
    )

    if or_qs.filter(authors=request.user).count() == 1:
        if or_qs[0].counter == 1:
            or_qs[0].delete_from_cache()
            or_qs.delete()
        else:
            or_qs.update(counter=F("counter") - 1)
            or_qs[0].delete_from_cache()
            or_qs[0].authors.remove(request.user)
    else:
        obj_reaction, created = ObjectReaction.objects.get_or_create(
            reaction=request.POST["reaction"],
            content_type=ctype,
            object_pk=object_pk,
            site=site,
        )
        obj_reaction.authors.add(request.user)
        obj_reaction.counter += 1
        obj_reaction.save()
    signals.object_got_a_reaction.send(
        sender=object.__class__,
        object=object,
        reaction=request.POST["reaction"],
        created=created,
        request=request,
    )
    return created


def list_reacted(request, comment_id, reaction_value):
    comment = get_object_or_404(
        get_comment_model(),
        pk=comment_id,
        site__pk=utils.get_current_site_id(request),
    )
    reaction = get_object_or_404(
        CommentReaction, reaction=reaction_value, comment=comment
    )

    authors = [
        settings.COMMENTS_INK_API_USER_REPR(author)
        for author in reaction.authors.all()
    ]

    max_users_in_tooltip = settings.COMMENTS_INK_MAX_USERS_IN_TOOLTIP
    if len(authors) <= max_users_in_tooltip:
        raise Http404(_("Not enough users"))

    paginator = Paginator(authors, settings.COMMENTS_INK_USERS_REACTED_PER_PAGE)
    page_number = request.GET.get("page", 1)
    page_obj = paginator.get_page(page_number)

    return render(
        request,
        _list_reacted_tmpl,
        {
            "comment": comment,
            "reaction": get_comment_reactions_enum()(reaction.reaction),
            "page_obj": page_obj,
        },
    )


def list_reacted_to_object(request, content_type_id, object_pk, reaction_value):
    try:
        ctype = ContentType.objects.get(pk=content_type_id)
        ctype.get_object_for_this_type(pk=object_pk)
    except Exception:
        raise Http404(
            "Object referenced by pair (ctype_id, obj_id): (%d, %d) "
            "does not exist" % (content_type_id, object_pk)
        )

    reaction = get_object_or_404(
        ObjectReaction,
        reaction=reaction_value,
        content_type=ctype,
        object_pk=object_pk,
        site=get_current_site(request),
    )

    authors = [
        settings.COMMENTS_INK_API_USER_REPR(author)
        for author in reaction.authors.all()
    ]

    max_users_in_tooltip = settings.COMMENTS_INK_MAX_USERS_IN_TOOLTIP
    if len(authors) <= max_users_in_tooltip:
        raise Http404(_("Not enough users"))

    paginator = Paginator(authors, settings.COMMENTS_INK_USERS_REACTED_PER_PAGE)
    page_number = request.GET.get("page", 1)
    page_obj = paginator.get_page(page_number)

    return render(
        request,
        _list_reacted_to_object_tmpl,
        {
            "content_type": ctype,
            "object": reaction.content_object,
            "reaction": get_object_reactions_enum()(reaction.reaction),
            "page_obj": page_obj,
        },
    )


def get_inkcomment_url(request, content_type_id, object_id, comment_id):
    response = shortcut(request, content_type_id, object_id)

    cpage_param = settings.COMMENTS_INK_PAGE_QUERY_STRING_PARAM
    cpage = request and request.GET.get(cpage_param, None) or None
    cfold_param = settings.COMMENTS_INK_FOLD_QUERY_STRING_PARAM
    cfold = request and request.GET.get(cfold_param, None) or ""

    if not cpage:
        # Create a CommentsPaginator and get the page of the comment.
        cpage = utils.get_comment_page_number(
            request, content_type_id, object_id, comment_id
        )

    qs_params = []

    try:
        page_number = int(cpage)
        if page_number > 1:
            qs_params.append(f"{cpage_param}={page_number}")
    except ValueError:
        if cpage == "last":
            qs_params.append(f"{cpage_param}={cpage}")
        else:
            raise Http404(
                _("Page is not “last”, nor can it be converted to an int.")
            )

    try:
        if len(cfold):
            {int(cid) for cid in cfold.split(",")}
            qs_params.append(f"{cfold_param}={cfold}")
    except (TypeError, ValueError):
        raise Http404(
            _("A comment ID in the list of folded comments is not an integer.")
        )

    if len(qs_params):
        return HttpResponseRedirect(f"{response.url}?{'&'.join(qs_params)}")
    else:
        return HttpResponseRedirect(response.url)
